"""
Utility functions for the simeon CLI tool
"""
import argparse
import configparser
import glob
import logging
import os
import shlex
import socket
import stat
import subprocess as sb
import sys
import tempfile
from argparse import ArgumentTypeError

from dateutil.parser import parse as dateparse


CLI_MAIN_EPILOG = """
RETURN CODES:
    simeon returns either 0 or 1, depending on whether an error was encountered.
    If any error is encountered with any of the subcommands, 1 is returned.
    For simeon list and simeon download, if nothing is listed or downloaded, then 1 is returned.
    For simeon split and simeon push, if nothing ends up being split or pushed, then 1 is returned.
    For simeon report, if any error is encountered while running the queries, then 1 is returned.

SETUP and CONFIGURATIONS:
    simeon is a glorified downloader and uploader set of scripts. Much of the downloading and uploading that it does makes the assumptions that you have
    your AWS credentials configured properly and that you've got a service account file for GCP services available on your machine. If the latter is
    missing, you may have to authenticate to GCP services through the SDK. However, both we and Google recommend the use of service accounts.

    Every downloaded file is decrypted either during the download process or while it gets split by the simeon split command. So, this tool assumes that
    you have installed and configured gpg to be able to decrypt files from edX.

    The following steps may be useful to someone just getting started with the edX data package:

    1. Credentials from edX

        o Reach out to edX to get your data czar credentials

        o Configure both AWS and gpg, so your credentials can access the S3 buckets and your gpg key can decrypt the files there

    2. Setup a GCP project

        o Create a GCP project

        o Setup a BigQuery workspace

        o Create a GCS bucket

        o Create a service account and download the associated file

        o Give the service account Admin Role access to both the BigQuery project and the GCS bucket

    If the above steps are carried out successfully, then you should be able to use simeon without any issues.

    However, if you have taken care of the above steps but are still unable to get simeon to work, please open an issue.

    Further, simeon can parse INI formatted configuration files. It, by default, looks for files in the user's home directory, or in the current working
    directory of the running process. The base names that are targeted when config files are looked up are: simeon.cfg or .simeon.cfg or simeon.ini or .simeon.ini.
    You can also provide simeon with a config file by using the global option --config-file or -C, and giving it a path to the file with the corresponding configurations.

    The following is a sample file content:

        # Default section for things like the organization whose data package is processed
        # You can also set a default site as one of the following: edx, edge, patches
        [DEFAULT]
        site = edx
        org = yourorganizationx
        clistings_file = /path/to/file/with/course_ids

        # Section related to Google Cloud (project, bucket, service account)
        [GCP]
        project = your-gcp-project-id
        bucket = your-gcs-bucket
        service_account_file = /path/to/a/service_account_file.json
        wait_for_loads = True
        geo_table = your-gcp-project.geocode_latest.geoip
        youtube_table = your-gcp-project.videos.youtube
        youtube_token = your-YouTube-API-token

        # Section related to the AWS credentials needed to download data from S3
        [AWS]
        aws_cred_file = ~/.aws/credentials
        profile_name = default

    The options in the config file(s) should match the optional arguments of the CLI tool. For instance, the --service-account-file, --project and
    --bucket options can be provided under the GCP section of the config file as service_account_file, project and bucket, respectively. Similarly, the
    --site and --org options can be provided under the DEFAULT section as site and org, respectively.


EXAMPLES:
List files
    simeon can list files on S3 for your organization based on criteria like file type (sql or log or email), time intervals (begin and end dates),
    and site (edx or edge or patches).
        # List the latest SQL data dump
        simeon list -s edx -o mitx -f sql -L
        # List the latest email data dump
        simeon list -s edx -o mitx -f email -L
        # List the latest tracking log file
        simeon list -s edx -o mitx -f log -L

Download and split files
    simeon can download, decrypt and split up files into folders belonging to specific courses.

    o Example 1: Download, split and push SQL bundles to both GCS and BigQuery

        # Download the latest SQL data dump
        simeon download -s edx -o mitx -f sql -L -d data/

        # Download SQL bundles dumped any time since 2021-01-01 and
        # extract the contents for course ID MITx/12.3x/1T2021.
        # Place the downloaded files in data/ and the output of the split in data/SQL
        simeon download -s edx -o mitx -c "MITx/12.3x/1T2021" -f sql -b 2021-01-01 -d data -S -D data/SQL/

        # Push to GCS the split up SQL files inside data/SQL/MITx__12_3x__1T2021
        simeon push gcs -f sql -p ${GCP_PROJECT_ID} -b ${GCS_BUCKET} -S ${SAFILE} data/SQL/MITx__12_3x__1T2021

        # Push the files to BigQuery and wait for the jobs to finish
        # Using -s or --use-storage tells BigQuery to extract the files
        # to be loaded from Google Cloud Storage.
        # So, use the option when you've already called simeon push gcs
        simeon push bq -w -s -f sql -p ${GCP_PROJECT_ID} -b ${GCS_BUCKET} -S ${SAFILE} data/SQL/MITx__12_3x__1T2021

    o Example 2: Download, split and push tracking logs to both GCS and BigQuery

        # Download the latest tracking log file
        simeon download -s edx -o mitx -f log -L -d data/

        # Download tracking logs dumped any time since 2021-01-01
        # and extract the contents for course ID MITx/12.3x/1T2021
        # Place the downloaded files in data/ and the output of the split in data/TRACKING_LOGS
        simeon download -s edx -o mitx -c "MITx/12.3x/1T2021" -f log -b 2021-01-01 -d data -S -D data/TRACKING_LOGS/

        # Push to GCS the split up tracking log files inside
        # data/TRACKING_LOGS/MITx__12_3x__1T2021
        simeon push gcs -f log -p ${GCP_PROJECT_ID} -b ${GCS_BUCKET} -S ${SAFILE} data/TRACKING_LOGS/MITx__12_3x__1T2021

        # Push the files to BigQuery and wait for the jobs to finish
        # Using -s or --use-storage tells BigQuery to extract the files
        # to be loaded from Google Cloud Storage.
        # So, use the option when you've already called simeon push gcs
        simeon push bq -w -s -f log -p ${GCP_PROJECT_ID} -b ${GCS_BUCKET} -S ${SAFILE} data/TRACKING_LOGS/MITx__12_3x__1T2021

    o If you have already downloaded SQL bundles or tracking log files, you can use simeon split them up.

Make secondary/aggregated tables
    simeon can generate secondary tables based on already loaded data. Call simeon report --help for the expected positional and optional arguments.

    o Example: Make person_course for course ID MITx/12.3x/1T2021

        # Make a person course table for course ID MITx/12.3x/1T2021
        # Provide the -g option to give a geolocation BigQuery table
        # to fill the ip-to-location details in the generated person course table
        COURSE=MITx/12.3x/1T2021
        simeon report -w -g "${GCP_PROJECT_ID}.geocode.geoip" -t "person_course" -p ${GCP_PROJECT_ID} -S ${SAFILE} ${COURSE}


NOTES:
1. Please note that SQL bundles are quite large when split up, so consider using the -c or --courses option when invoking simeon download -S or
    simeon split to make sure that you limit the splitting to a set of course IDs. The `--clistings-file` option is an alternative to `--courses`.
    It expects a text file with one course ID per line.
    If those options are not used, simeon may end up failing to complete the split operation
    due to exhausted system resources (storage to be specific).

2. simeon download with file types log and email will both download and decrypt the files matching the given criteria. If the latter operations are
    successful, then the encrypted files are deleted by default. This is to make sure that you don't exhaust storage resources. If you wish to keep
    those files, you can always use the --keep-encrypted option that comes with simeon download and simeon split. SQL bundles are only downloaded (not decrypted).
    Their decryption is done during a split operation.

3. Unless there is an unhandled exception (which should be reported as a bug), simeon should, by default, print to the standard output both information
    and errors encountered while processing your files. You can capture those logs in a file by using the global option --log-file and providing
    a destination file for the logs.

4. When using multi argument options like --tables or --courses, you should try not to place them right before the expected positional arguments.
    This will help the CLI parser not confuse your positional arguments with table names (in the case of --tables) or course IDs (when --courses is used).

5. Splitting tracking logs is a resource intensive process. The routine that splits the logs generates a file for each course ID encountered. If you
    happen to have more course IDs in your logs than the running process can open operating system file descriptors, then simeon will put away records
    it cannot save to disk for a second pass. Putting away the records involves using more memory than normally required. The second pass will only
    require one file descriptor at a time, so it should be safe in terms of file descriptor limits. To help simeon not have to do a second pass, you
    may increase the file descriptor limits of processes from your shell by running something like ulimit -n 2000 before calling simeon split on Unix
    machines. For Windows users, you may have to dig into the Windows Registries for a corresponding setting. This should tell your OS kernel to allow
    OS processes to open up to 2000 file handles.

6. Care must be taken when using simeon split and simeon push to make sure that the number of positional arguments passed does not lead to the
    invoked command exceeding the maximum command-line length allowed for arguments in a command. To avoid errors along those lines, please consider
    passing the positional arguments as UNIX glob patterns. For instance, simeon split --file-type log 'data/TRACKING-LOGS/*/*.log.gz' tells simeon to
    expand the given glob pattern, instead of relying on the shell to do it.

7. The report subcommand relies on the presence of SQL query files to parse and send to BigQuery to execute. Any errors arising from executing the parsed
    queries will be shown to the end user through the given log stream. While the simeon tool ships with query files for most secondary/reporting tables
    that are based on the edx2bigquery tool, an end user should be able to point simeon to a different location with SQL query files by using
    the --query-dir option that comes with simeon report. Additionally, these query files can contain jinja2 templated SQL code.
    Any mentioned variables within these templated queries can be passed to simeon report by using the --extra-args option and passing key-value pair items
    in the format var1=value1,var2=value2,var3=value3,...,varn=valuen. Further, these key-value pair items can also be typed by using the format
    var1:i=value1,var2:s=value2,var3:f=value3,...,varn:s=valuen. In this format, the type is append to the key, separated by a colon.
    The only supported scalar types, so far, are s for str, i for int, and f for float. If any conversion errors occur during value parsing,
    then those are shown to the end user, and the query won't get executed. Finally, if you wish to pass an array or list to the template,
    you will need to repeat a key multiple times. For instance, if you want to pass a list named mylist containing the integers,
    you could write something like --extra-args mylist:i=1,mylist:i=2,mylist:i=3. This means that you'll have a python list named
    mylist within your template, and it should contain [1, 2, 3].
"""
CONFIGS = {
    'DEFAULT': (
        ('site', configparser.ConfigParser.get),
        ('org', configparser.ConfigParser.get),
        ('clistings_file', configparser.ConfigParser.get),
        ('youtube_token', configparser.ConfigParser.get),
        ('file_format', configparser.ConfigParser.get),
        ('schema_dir', configparser.ConfigParser.get),
        ('max_bad_rows', configparser.ConfigParser.getint),
        ('update_description', configparser.ConfigParser.getboolean),
        ('schema_dir', configparser.ConfigParser.get),
        ('query_dir', configparser.ConfigParser.get),
        ('project', configparser.ConfigParser.get),
        ('bucket', configparser.ConfigParser.get),
        ('service_account_file', configparser.ConfigParser.get),
        ('geo_table', configparser.ConfigParser.get),
        ('youtube_table', configparser.ConfigParser.get),
        ('extra_args', configparser.ConfigParser.get),
        ('email_bucket', configparser.ConfigParser.get),
        ('sql_bucket', configparser.ConfigParser.get),
        ('log_bucket', configparser.ConfigParser.get),
        ('rdx_bucket', configparser.ConfigParser.get),
    ),
    'GCP': (
        ('project', configparser.ConfigParser.get),
        ('bucket', configparser.ConfigParser.get),
        ('service_account_file', configparser.ConfigParser.get),
        ('wait_for_loads', configparser.ConfigParser.getboolean),
        ('use_storage', configparser.ConfigParser.getboolean),
        ('geo_table', configparser.ConfigParser.get),
        ('youtube_table', configparser.ConfigParser.get),
        ('youtube_token', configparser.ConfigParser.get),
        ('file_format', configparser.ConfigParser.get),
        ('schema_dir', configparser.ConfigParser.get),
        ('query_dir', configparser.ConfigParser.get),
        ('max_bad_rows', configparser.ConfigParser.getint),
        ('extra_args', configparser.ConfigParser.get),
    ),
    'AWS': (
        ('aws_cred_file', configparser.ConfigParser.get),
        ('profile_name', configparser.ConfigParser.get),
        ('email_bucket', configparser.ConfigParser.get),
        ('sql_bucket', configparser.ConfigParser.get),
        ('log_bucket', configparser.ConfigParser.get),
        ('rdx_bucket', configparser.ConfigParser.get),
    ),
}
REPORT_TABLES = [
    'video_axis', 'forum_events', 'problem_grades', 'chapter_grades',
    'show_answer', 'video_stats_day', 'show_answer_stats_by_user',
    'show_answer_stats_by_course', 'course_item', 'person_item',
    'person_problem', 'course_problem', 'person_course_day',
    'pc_video_watched', 'pc_day_totals', 'pc_day_trlang',
    'pc_day_ip_counts', 'language_multi_transcripts', 'pc_nchapters',
    'pc_forum', 'course_modal_language', 'course_modal_ip',
    'forum_posts', 'forum_person', 'enrollment_events', 'enrollday_all',
    'person_enrollment_verified', 'pc_day_agent_counts', 'course_modal_agent',
    'person_course',
]
EXTRA_ARG_TYPE = {'i': int, 'f': float, 's': str,}


def parsed_date(datestr: str) -> str:
    """
    Function to parse the --start-date and --end-date
    options of simeon

    :type datestr: str
    :param datestr: A stringified date
    :rtype: str
    :returns: A properly formatted date string
    :raises: ArgumentTypeError
    """
    try:
        return dateparse(datestr).strftime('%Y-%m-%d')
    except Exception:
        msg = '{d!r} could not be parsed into a proper date'
        raise ArgumentTypeError(msg.format(d=datestr))


def gcs_bucket(bucket: str) -> str:
    """
    Clean up a GCS bucket name if it does not start with the gs:// protocol

    :type bucket: str
    :param bucket: Google Cloud Storage bucket name
    :rtype: str
    :returns: A properly formatted GCS bucket name
    """
    if not bucket.startswith('gs://'):
        return 'gs://{b}'.format(b=bucket)
    return bucket


def bq_table(name):
    """
    Check that the given BigQuery table name
    is a fully qualified one: dataset.table or project.dataset.table
    """
    if not name:
        return None
    chunks = name.split('.')
    if len(chunks) not in (2, 3):
        raise ArgumentTypeError(
            '{n} is not a valid BigQuery table name.\nValid table names '
            'should be in the form project.dataset.table '
            'or dataset.table.'.format(n=name)
        )
    return name


def optional_file(fname: str) -> str:
    """
    Clean up a given a file path if it's not None.
    Also, check that it exists. Otherwise, raise ArgumentTypeError

    :type fname: str
    :param fname: File name from the command line
    :rtype: str
    :returns: A properly formatted file name
    :raises: ArgumentTypeError
    """
    if fname is not None:
        fname = os.path.expanduser(fname)
        if not os.path.exists(fname):
            msg = 'The given file name {f!r} does not exist.'
            raise ArgumentTypeError(msg.format(f=fname))
    return os.path.realpath(fname)


def make_logger(user='SIMEON', verbose=True, stream=None):
    """
    Create a Logger object pointing to the given stream

    :type verbose: bool
    :param verbose: If True, log level is INFO. Otherwise, it's WARN
    :type stream: Union[TextIOWrapper,None]
    :param stream: A file object opened for writing
    :rtype: logging.Logger
    :returns: Returns a Logger object used to print messages
    """
    if stream is None:
        stream = sys.stdout
    if not hasattr(stream, 'write'):
        stream = open(stream, 'a')
    level = logging.INFO if verbose else logging.WARN
    formatter = logging.Formatter(
        '%(asctime)s:%(hostname)s:%(levelname)s:%(name)s:%(message)s',
        '%Y-%m-%d %H:%M:%S%z'
    )
    logger = logging.getLogger(user)
    handler = logging.StreamHandler(stream=stream)
    handler.setLevel(level)
    handler.set_name(user)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logging.LoggerAdapter(logger, {'hostname': socket.gethostname()})


def make_config_file(output=None):
    """
    Create a config file named 'simeon.cfg' that will have the expected
    configuration values

    :type output: Union[None, str, pathlib.Path]
    :param output: Path to the config file where sections and options are put
    :rtype: None
    :returns: Write the config info to the given output file
    """
    if output is None:
        output = os.path.join(os.path.expanduser('~'), 'simeon.cfg')
    config = configparser.ConfigParser()
    config.optionxform = lambda s: s.lstrip('-').lower().replace('-', '_')
    config['DEFAULT'] = {
        'site': '',
        'org': '',
    }
    config['GCP'] = {
        'project': '',
        'bucket': '',
        'service_account_file': '',
    }
    config['AWS'] = {
        'credential_file': '',
        'profile_name': '',
    }
    with open(output, 'w') as configfile:
        config.write(configfile)


def find_config(fname=None, no_raise=False):
    """
    Searches for config files in default locations.
    If no file name is provided, it tries to load files
    in the home and current directories of the running process.

    :type fname: Union[None, str, pathlib.Path]
    :param fname: Path to an INI config file, default "simeon.cfg"
    :rtype: configparser.ConfigParser
    :returns: Returns a ConfigParser with configs from the file(s)
    """
    if fname is None:
        names = ('simeon.cfg', 'simeon.ini', '.simeon.cfg', '.simeon.ini')
        files = []
        for name in names:
            files.extend([
                os.path.join(os.path.expanduser('~'), name),
                os.path.join(os.getcwd(), name)
            ])
    else:
        files = [os.path.expanduser(fname)]
    config = configparser.ConfigParser()
    config.optionxform = lambda s: s.lstrip('-').lower().replace('-', '_')
    for config_file in files:
        try:
            config.read(config_file)
        except Exception as excp:
            if no_raise:
                continue
            raise ArgumentTypeError(excp) from None
    return config


def course_listings(courses):
    """
    Given a list of white space separated course IDs,
    split it into a list.

    :type courses: Union[Iterable, None]
    :param courses: An iterable of course ID's
    :rtype: set
    :returns: A set object of course IDs
    """
    if courses is None:
        return None
    out = set()
    for c in courses:
        out.add(c.strip().split(':')[-1].replace('+', '/'))
    return out


def courses_from_file(fname):
    """
    Given the path to a file, extract the course IDs in it.
    One course ID per line.

    :type fname: str
    :param fname: Path to a file with course listings. 1 course ID per line
    :rtype: Union[None, set]
    :returns: A set of course IDs
    """
    if not fname:
        return None
    fname = os.path.expanduser(fname)
    if not os.path.isfile(fname):
        msg = 'The given course listings file, {f!r}, is not a valid file'
        raise ArgumentTypeError(msg.format(f=fname))
    out = set()
    with open(fname) as fh:
        for c in fh:
            out.add(c.strip().split(':')[-1].replace('+', '/'))
    return out


def course_paths_from_file(fname):
    """
    Given the path to a file, extract the course IDs in it, and
    turn the course IDs into matching directory names.

    :type fname: str
    :param fname: Path to a file with course listings. 1 course ID per line
    :rtype: Union[None, set]
    :returns: A set of directory names
    """
    if not fname:
        return None
    fname = os.path.expanduser(fname)
    if not os.path.isfile(fname):
        msg = 'The given course listings file, {f!r}, is not a valid file'
        raise ArgumentTypeError(msg.format(f=fname))
    out = set()
    with open(fname) as fh:
        for line in fh:
            line = line.split(':')[-1]
            c = line.strip().replace('/', '__').replace('+', '__')
            out.add(c.replace('.', '_').replace('-', '_').lower())
        return out


def filter_generated_items(items, cdirs):
    """
    This is used to filter the given items (which are file paths)
    using the provided course ID directories. Since the directories
    are generated by the course_paths_from_file function, the items
    in cdirs are expected to be lowercased.

    :type items: Union[List[str], Set[str]]
    :param items: Items passed to the simeon push command
    :type cdirs: Set[str]
    :param cdirs: Course ID directories (lowercased)
    :rtype: Set[str]
    :return: Returns a set of file paths whose directories are in cdirs
    """
    if not cdirs:
        return set(items)
    out = set()
    for item in items:
        if '*' not in item:
            if os.path.isdir(item):
                real_item = os.path.basename(os.path.realpath(item))
            else:
                real_item = os.path.basename(
                    os.path.dirname(os.path.realpath(item))
                )
            if real_item.lower() in cdirs:
                out.add(item)
        else:
            out.update(filter_generated_items(glob.glob(item), cdirs))
    return out


def items_from_files(files):
    """
    Given a list of text files, extract file paths
    """
    out = set()
    for file_ in files:
        with open(file_) as fh:
            out.update(map(str.strip, fh))
    return out


def expand_paths(items):
    """
    Expand glob patterns in items
    """
    out = []
    for path in items:
        out.extend(glob.iglob(path))
    return out


def is_parallel(args):
    """
    Take a parsed argparse.Namespace and check that simeon
    will be using some paraellism.
    It also sets an is_parallel on the Namespace object
    """
    if getattr(args, 'command', '') not in ('download', 'split', 'report'):
        args.is_parallel = False
    elif getattr(args, 'dynamic_date', False):
        args.is_parallel = False
    elif hasattr(args, 'in_files'):
        args.is_parallel = True
    else:
        items = getattr(
            args, 'downloaded_files', getattr(args, 'course_ids', [])
        )
        args.is_parallel = len(expand_paths(items)) > 1
    return args.is_parallel


def process_extra_args(extras):
    """
    Take a string in the form "var1=val1,var2,val2,...,varn=valn" and convert
    it to a dict

    :type extras: str
    :param extras: String in the form "var1=val1,var2,val2,...,varn=valn"
    :rtype: dict
    :return: Returns a dict from the string: var for key and val for value
    """
    if not extras:
        return {}
    out = dict()
    types = EXTRA_ARG_TYPE
    extras = map(lambda p: p.lstrip().split('=')[:2], extras.split(','))
    while 1:
        try:
            k, v = next(extras)
            chunks = iter(k.split(':'))
            k = next(chunks)
            type_ = next(chunks, None)
            func = types.get(type_) or str
            v = func(v)
            if k in out:
                if not isinstance(out[k], list):
                    out[k] = [out[k]]
                out[k].append(v)
            else:
                out[k] = v
        except StopIteration:
            break
        except ValueError as excp:
            msg = (
                'The provided extra arguments are not properly formatted: {e}'
            )
            raise Exception(msg.format(e=excp)) from None
    return out


class NumberRange(object):
    """
    Check if a number is within a range. Otherwise, raise ArgumentTypeError
    """
    def __init__(self, type_=int, lower=1, upper=50):
        self.type_ = type_
        self.lower = lower
        self.upper = upper

    def __call__(self, value):
        try:
            v = self.type_(value)
        except Exception as excp:
            msg = (
                'The given value {v} could not be converted to a number. '
                'Please provide a valid value: {e}.'
            )
            raise ArgumentTypeError(msg.format(v=value, e=excp)) from None
        if not (self.lower <= v <= self.upper):
            msg = '{v} is not in the range [{l}, {u}]'
            raise ArgumentTypeError(
                msg.format(v=value, l=self.lower, u=self.upper)
            )
        return v


class TableOrderAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string):
        if any(k in sys.argv for k in ('--no-reorder', '-O')):
            print(values)
            setattr(namespace, 'tables', values)
            return
        tables = []
        for table in REPORT_TABLES:
            if table in values:
                tables.append(table)
        if not tables:
            if not values:
                raise ArgumentTypeError(
                    'Table names required when --tables is given'
                )
            setattr(namespace, 'tables', values)
            return
        for table in values:
            if table not in tables:
                tables.append(table)
        setattr(namespace, 'tables', tables)


def get_pager():
    """
    Get path to less or more
    """
    pagers = (os.getenv('PAGER'), 'less', 'more',)
    for path in (os.getenv('PATH') or '').split(os.path.pathsep):
        for pager in pagers:
            if pager is None:
                continue
            pager = iter(pager.split(' ', 1))
            prog = os.path.join(path, next(pager))
            args = next(pager, None) or ''
            try:
                md = os.stat(prog).st_mode
                if md & (stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH):
                    return '{p} {a}'.format(p=prog, a=args)
            except OSError:
                continue


class CustomArgParser(argparse.ArgumentParser):
    """
    A custom ArgumentParser class that prints help messages
    using either less or more, if available. Otherwise, it does
    what ArgumentParser does.
    """
    def print_help(self, file=None):
        text = self.format_help()
        pager = get_pager()
        if pager is None:
            return super().print_help(file)
        fd, fname = tempfile.mkstemp(prefix='simeon_help_', suffix='.txt')
        with open(fd, 'w') as fh:
            super().print_help(fh)
        cmd = shlex.split('{p} {f}'.format(p=pager, f=fname))
        with sb.Popen(cmd) as proc:
            rc = proc.wait()
            try:
                os.unlink(fname)
            except:
                pass
            if rc != 0:
                return super().print_help(file)


__all__ = [
    'CustomArgParser', 'NumberRange', 'TableOrderAction', 'bq_table',
    'course_listings', 'course_paths_from_file', 'courses_from_file',
    'expand_paths', 'filter_generated_items', 'find_config', 'gcs_bucket',
    'get_pager', 'is_parallel', 'items_from_files', 'make_config_file',
    'make_logger', 'optional_file', 'parsed_date', 'process_extra_args',
]



"""
simeon is a python package and CLI tool that helps with fetching and processing edX research data.
"""
import logging


__version__ = '0.0.21'

# Set loggers for the scripts and package users
# 1. for simeon
logging.getLogger('SIMEON')
# 2. simeon-geoip
logging.getLogger('SIMEON-GEOIP')
# 3. simeon-youtube
logging.getLogger('SIMEON-YOUTUBE')



"""
Unit tests for the upload package
"""
import unittest

from simeon.upload import utilities as uputils
from simeon.exceptions import (
    BigQueryNameException, MissingSchemaException
)


class TestUploadUtilities(unittest.TestCase):
    """
    Test the utility functions and classes in the upload package
    """
    def setUp(self):
        self.good_file_mames = [
            ('log', 'data/MITx__123__2T2020/tracklog-2020-06-01.json.gz'),
            ('sql', 'data/MITx__123__2T2020/user_info_combo.csv.gz'),
            ('sql', 'data/MITx__123__2T2020/person_course.csv.gz'),
            ('sql', 'data/MITx__123__2T2020/forum.csv.gz'),
            ('sql', 'data/MITx__123__2T2020/forum_person.csv.gz'),
            ('email', 'data/MITx__123__2T2020/email-opt-in.csv.gz'),
        ]
        self.bad_file_names = [
            ('log', 'data/MITx__123__2T2020/no-date-here.json.gz'),
            ('sql', 'data/MITx__123__2T2020/.hidden.csv.gz')
        ]
        self.bad_file_types = [
            ('bad', 'data/MITx__123__2T2020/tracklog-2020-06-01.json.gz'),
            ('worse', 'data/MITx__123__2T2020/user_info_combo.csv.gz'),
            ('worst', 'data/MITx__123__2T2020/person_course.csv.gz'),
            ('bad', 'data/MITx__123__2T2020/forum.csv.gz'),
            ('worse', 'data/MITx__123__2T2020/forum_person.csv.gz'),
            ('worst', 'data/MITx__123__2T2020/email-opt-in.csv.gz'),
        ]
        self.missing_schema_tables = [
            'this_table_doesnt_exist_anywhere',
            'neither_does_this',
        ]
        self.project = 'our-awesome-project'
        self.bucket = 'our-awesome-bucket'
    
    def test_good_table_names(self):
        """
        Test valid BigQuery table names from file names
        """
        for ftype, fname in self.good_file_mames:
            msg = 'With file name {f} and type {t}'.format(f=fname, t=ftype)
            with self.subTest(msg):
                out = uputils.local_to_bq_table(fname, ftype, self.project)
                self.assertTrue(len(out.split('.')) == 3)
    
    def test_bad_table_names(self):
        """
        Test that badly named data files can't make BigQuery table names
        """
        for ftype, fname in self.bad_file_names:
            msg = 'With file name {f} and type {t}'.format(f=fname, t=ftype)
            with self.subTest(msg):
                with self.assertRaises(BigQueryNameException):
                    uputils.local_to_bq_table(fname, ftype, self.project)
    

    def test_missing_schema(self):
        """
        Test that calling uputils.get_bq_schema for a table with missing
        schema file causes MissingSchemaException
        """
        for table in self.missing_schema_tables:
            with self.assertRaises(MissingSchemaException):
                uputils.get_bq_schema(table)

    def test_local_to_bq_table_with_unknown_file_types(self):
        """
        Test that uputils.local_to_bq_table raises a ValueError when
        given an unknown file type.
        """
        for ftype, fname in self.bad_file_types:
            msg = 'With file name {f} and ftype {t}'.format(f=fname, t=ftype)
            with self.subTest(msg):
                with self.assertRaises(ValueError):
                    uputils.local_to_bq_table(fname, ftype, self.project)


if __name__ == '__main__':
    unittest.main()



"""
Utilities functions and classes to help with loading data to Google Cloud
"""
import glob
import gzip
import os
from jinja2 import Template
import uuid
from datetime import datetime
from typing import List

from google.cloud import (
    bigquery, exceptions as gcp_exceptions, storage
)

from simeon.upload import utilities as uputils
from simeon.report import utilities as rutils
from simeon.exceptions import LoadJobException
from simeon.upload.utilities import (
    SCHEMA_DIR, course_to_bq_dataset
)


FILE_FORMATS = {
    'log': ['json'],
    'sql': ['csv', 'txt', 'sql', 'json']
}
MERGE_DDL = """MERGE {{ first }} f USING {{ second }} s
ON f.{{ column }} = s.{{ column }}
WHEN NOT MATCHED THEN
INSERT ROW
{%- if update_cols is defined and update_cols %}
WHEN MATCHED {% if match_equal_columns is defined and match_equal_columns -%}
AND {% for col in match_equal_columns %}
s.{{ col }} = f.{{ col }} {% if not loop.last %} AND {% endif %}
{% endfor %}
{% endif %}
{% if match_unequal_columns is defined and match_unequal_columns -%}
AND {% for col in match_unequal_columns %}
s.{{ col }} <> f.{{ col }} {% if not loop.last %} AND {% endif %}
{% endfor -%}
{% endif -%}
THEN UPDATE SET {% for col in update_cols -%}
    f.{{ col }} = s.{{ col }}{% if not loop.last %}, {% endif %}
{%- endfor %}
{% endif %}
"""
DST_DESC = {
    'log': 'Dataset to host the tracking log data from edX courses',
    'sql': (
        'Dataset to host all the tables that are computed from a combination'
        ' of tab-delimited files from edX\'s weekly SQL data dump and '
        'tracking log tables.'
    ),
    'email': (
        'Dataset to host the dimensional details about users\' email'
        ' adresses and email opt-in preferences'
    ),
}


class BigqueryClient(bigquery.Client):
    """
    Subclass bigquery.Client and add convenience methods
    """
    def get_course_tables(self, course_id):
        """
        Get all the tables related to the given course ID

        :type course_id: str
        :param course_id: edX course ID in format ORG/NUMBER/TERM
        :rtype: Dict[str, set]
        :return: A dict with keys as log and sql, and values as table names
        """
        out = {'log': set(), 'sql': set()}
        log_dset = course_to_bq_dataset(course_id, 'log', self.project)
        latest_dset = course_to_bq_dataset(course_id, 'sql', self.project)
        query = """select table_schema, table_name
        from `region-us`.INFORMATION_SCHEMA.TABLES
        where table_schema in ('{f}', '{s}')"""
        query = query.format(
            f=log_dset.split('.')[-1], s=latest_dset.split('.')[-1]
        )
        job = self.query(query)
        for row in job.result():
            ds = row.get('table_schema')
            if ds.endswith('_logs'):
                out['log'].add(row.get('table_name'))
            else:
                out['sql'].add(row.get('table_name'))
        return out

    def load_tables_from_dir(
        self, dirname: str, file_type: str, project: str,
        create: bool, append: bool, use_storage: bool=False,
        bucket: str=None, max_bad_rows=0,
        schema_dir=SCHEMA_DIR, format_='json', patch=False,
    ) -> List[bigquery.LoadJob]:
        """
        Load all the files in the given directory.

        :type dirname: str
        :param dirname: Grandparent or parent directory of split up files
        :type file_type: str
        :param file_type: One of sql, email, log, rdx
        :type project: str
        :param project: Target GCP project
        :type create: bool
        :param create: Whether or not to create the destination table
        :type append: bool
        :param append: Whether or not to append the records to the table
        :type use_storage: bool
        :param use_storage: Whether or not to load the data from GCS
        :type bucket: str
        :param bucket: GCS bucket name to use
        :type max_bad_rows: int
        :param max_bad_rows: Max number of bad rows allowed during loading
        :type schema_dir: str
        :param schema_dir: Directory where schema files are found
        :type format_: str
        :param format_: File format (json or csv)
        :type patch: bool
        :param patch: Whether or not to patch the description of the table
        :rtype: List[bigquery.LoadJob]
        :returns: List of load jobs
        :raises: Propagates everything from the underlying package
        """
        schema_dir = schema_dir or SCHEMA_DIR
        formats = FILE_FORMATS.get(file_type, [])
        patts = (
            os.path.join(dirname, '*.{f}*.gz'),
            os.path.join(dirname, '*', '*.{f}*.gz')
        )
        files = []
        for patt in patts:
            for format_ in formats:
                files.extend(glob.glob(patt.format(f=format_)))
        jobs = []
        for file_ in files:
            jobs.append(
                self.load_one_file_to_table(
                    file_, file_type, project, create, append,
                    use_storage, bucket, max_bad_rows, schema_dir, format_,
                    patch
                )
            )
        return jobs

    def load_one_file_to_table(
        self, fname: str, file_type: str, project: str,
        create: bool, append: bool, use_storage: bool=False,
        bucket: str=None, max_bad_rows=0,
        schema_dir=SCHEMA_DIR, format_='json', patch=False,
    ):
        """
        Load the given file to a target BigQuery table

        :type fname: str
        :param fname: The specific file to load
        :type file_type: str
        :param file_type: One of sql, email, log, rdx
        :type project: str
        :param project: Target GCP project
        :type create: bool
        :param create: Whether or not to create the destination table
        :type append: bool
        :param append: Whether or not to append the records to the table
        :type use_storage: bool
        :param use_storage: Whether or not to load the data from GCS
        :type bucket: str
        :param bucket: GCS bucket name to use
        :type max_bad_rows: int
        :param max_bad_rows: Max number of bad rows allowed during loading
        :type schema_dir: Union[None, str]
        :param schema_dir: Directory where schema files are found
        :type format_: str
        :param format_: File format (json or csv)
        :type patch: bool
        :param patch: Whether or not to patch the description of the table
        :rtype: bigquery.LoadJob
        :returns: The LoadJob object associated with the work being done
        :raises: Propagates everything from the underlying package
        """
        schema_dir = schema_dir or SCHEMA_DIR
        use_storage = use_storage or fname.startswith('gs://')
        if use_storage:
            if bucket is None:
                raise ValueError('use_storage=True requires a bucket name')
            loader = self.load_table_from_uri
        else:
            loader = self.load_table_from_file
        job_prefix = 'simeon_{t}_data_load_{dt}-'.format(
            t=file_type, dt=datetime.now().strftime('%Y%m%d%H%M%S%f')
        )
        dest = uputils.local_to_bq_table(fname, file_type, project)
        # dataset = self.dataset(dest.rsplit('.', 1)[0])
        dataset = bigquery.Dataset.from_string(dest.rsplit('.', 1)[0])
        dataset.description = DST_DESC.get(file_type)
        self.create_dataset(dataset, exists_ok=True)
        if use_storage:
            if not fname.startswith('gs://'):
                fname = uputils.local_to_gcs_path(fname, file_type, bucket)
        else:
            fname = gzip.open(fname, 'rb')
        config, desc = uputils.make_bq_load_config(
            table=dest, schema_dir=schema_dir,
            append=append, create=create, file_format=format_,
            max_bad_rows=max_bad_rows
        )
        dest = bigquery.Table.from_string(dest)
        dest.description = desc
        if patch:
            # Update the destination table's description.
            # Catch the error raised when the destination table
            # does not exist.
            try:
                self.update_table(dest, ['description'])
            except gcp_exceptions.NotFound:
                pass
        return loader(
            fname, dest, job_config=config, job_id_prefix=job_prefix
        )

    @staticmethod
    def extract_error_messages(errors):
        """
        Return the error messages from given list of error objects (dict)
        """
        messages = []
        if isinstance(errors, dict):
            members = []
            for e in errors.values():
                members += e
        else:
            members = errors
        for err in members:
            msg = err.get('message', '')
            if not msg:
                continue
            src = err.get('source', '')
            if src:
                msg = 'Source: {s} - {m}'.format(s=src, m=msg)
            loc = err.get('location', '')
            if loc:
                msg = '{m} - File: {f}'.format(m=msg, f=loc)
            messages.append(msg)
        return messages

    def merge_to_table(
        self, fname, table, col,
        schema_dir=SCHEMA_DIR, use_storage=False, patch=False,
        match_equal_columns=None, match_unequal_columns=None,
    ):
        """
        Merge the given file to the target table name.
        If the latter does not exist, create it first.
        This process waits for all the jobs needed

        :type fname: str
        :param fname: A local file name or a GCS URI
        :type table: str
        :param table: Fully qualified BigQuery table name
        :type col: str
        :param col: Column by which to merge
        :type schema_dir: Union[None, str]
        :param schema_dir: The directory where schema files live
        :type use_storage: bool
        :param use_storage: Whether or not the given path is a GCS URI
        :type patch: bool
        :param patch: Whether or not to patch the description of the table
        :type match_equal_columns: Union[List[str], None, Tuple[str]]
        :param match_equal_columns: List of column names for which to set
        equality (=) if WHEN MATCH is met during the merge.
        :type match_unequal_columns: Union[List[str], None, Tuple[str]]
        :param match_equal_columns: List of column names for which to set
        inequality (<>) if WHEN MATCH is met during the merge.
        :rtype: bigquery.QueryJob
        :returns: The QueryJob object associated with the merge carried out
        :raises: Propagates everything from the underlying package
        """
        schema_dir = schema_dir or SCHEMA_DIR
        if len(table.split('.')) < 3:
            table = '{p}.{t}'.format(p=self.project, t=table)
        dataset = table.rsplit('.', 1)[0]
        bqtable = bigquery.Table.from_string(table)
        temp_table_name = '{t}_{d}_{u}_temp'.format(
            t=table,
            d=datetime.now().strftime('%Y%m%d%H%M%S%f'),
            u=uuid.uuid4().hex.replace('-', '').replace('.', '')
        )
        temp_table = bigquery.Table.from_string(temp_table_name)
        schema, desc = uputils.get_bq_schema(table, schema_dir=schema_dir)
        bqtable.schema = schema
        bqtable.description = desc
        temp_table.schema = schema
        if patch:
            # Update the destination table's description.
            # Catch the error raised when the destination table
            # does not exist.
            try:
                self.update_table(bqtable, ['description'])
            except gcp_exceptions.NotFound:
                pass
        temp_table.description = desc
        job_prefix = 'simeon_{t}_data_load_{dt}-'.format(
            t=bqtable.table_id,
            dt=datetime.now().strftime('%Y%m%d%H%M%S%f')
        )
        config, _ = uputils.make_bq_load_config(
            table=table, schema_dir=schema_dir, append=False,
            create=True, file_format='json',
        )
        self.create_dataset(dataset, exists_ok=True)
        for tbl in (bqtable, temp_table):
            self.create_table(tbl, exists_ok=True)
        if use_storage:
            loader = self.load_table_from_uri
        else:
            loader = self.load_table_from_file
            fname = gzip.open(fname, 'rb')
        job = loader(
            fname, temp_table, job_config=config, job_id_prefix=job_prefix
        )
        rutils.wait_for_bq_job_ids([job.job_id], self)
        if job.errors:
            self.delete_table(temp_table, not_found_ok=True)
            msg = 'Merge job failed with: {e}'
            raise LoadJobException(msg.format(
                e='\n'.join(self.extract_error_messages(job.errors))
            ))
        query_template = Template(MERGE_DDL)
        update_cols = [f.name for f in bqtable.schema if f.name.lower() != col.lower()]
        query = query_template.render(
            first=table, second=temp_table_name,
            column=col, update_cols=update_cols,
            match_equal_columns=match_equal_columns,
            match_unequal_columns=match_unequal_columns,
        )
        qjob = self.query(query)
        rutils.wait_for_bq_job_ids([qjob.job_id], self)
        if qjob.errors:
            self.delete_table(temp_table, not_found_ok=True)
            msg = 'Merge job failed with: {e}'
            raise LoadJobException(msg.format(
                e='\n'.join(self.extract_error_messages(job.errors))
            ))
        self.delete_table(temp_table, not_found_ok=True)


class GCSClient(storage.Client):
    """
    Make a client to load data files to GCS
    """
    def load_one_file_to_gcs(
        self, fname: str, file_type: str, bucket: str
    ):
        """
        Load the given file to GCS

        :type fname: str
        :param fname: The local file to load to GCS
        :type file_type: str
        :param: file_type: One of sql, email, log, rdx
        :type bucket: str
        :param bucket: GCS bucket name
        :type overwrite: bool
        :param overwrite: Overwrite the target blob if it exists
        :rtype: None
        :returns: Nothing, but should load the given file to GCS
        :raises: Propagates everything from the underlying package
        """
        dest = storage.Blob.from_string(
            uputils.local_to_gcs_path(fname, file_type, bucket),
            client=self
        )
        if 'cold' in file_type.lower():
            dest.storage_class = storage.constants.COLDLINE_STORAGE_CLASS
        dest.upload_from_filename(fname, timeout=20 * 60)

    def load_dir(
        self, dirname: str, file_type: str, bucket: str
    ):
        """
        Load all the files in the given directory or
        any immediate subdirectories

        :type dirname: str
        :param dirname: The directory whose files are loaded
        :type file_type: str
        :param: file_type: One of sql, email, log, rdx
        :type bucket: str
        :param bucket: GCS bucket name
        :rtype: None
        :returns: Nothing, but should load file(s) in dirname to GCS
        :raises: Propagates everything from the underlying package
        """
        formats = FILE_FORMATS.get(file_type, [])
        patts = (
            os.path.join(dirname, '*.{f}*.gz'),
            os.path.join(dirname, '*', '*.{f}*.gz')
        )
        files = []
        for patt in patts:
            for format_ in formats:
                files.extend(glob.glob(patt.format(f=format_)))
        for fname in files:
            self.load_one_file_to_gcs(fname, file_type, bucket)



"""
simeon is a command line tool that helps with processing edx data
"""
import glob
import multiprocessing as mp
import os
import signal
import sys
import traceback
from argparse import (
    ArgumentParser, FileType, RawDescriptionHelpFormatter
)

import simeon
from simeon.download import (
    aws, emails, logs, sqls, utilities as downutils
)
from simeon.exceptions import EarlyExitError
from simeon.report import (
    QUERY_DIR, SCHEMA_DIR,
    make_sql_tables_par, make_sql_tables_seq,
    make_tables_from_sql, make_tables_from_sql_par, wait_for_bq_job_ids
)
from simeon.scripts import utilities as cli_utils
from simeon.upload import gcp

# Global logger variable for signal handlers
logger = None


def bail_out(sig, frame):
    """
    Exit somewhat cleanly from a signal
    """
    if logger:
        logger.warning('The process is being interrupted by a signal.')
    if logger:
        logger.warning('Waiting for child processes...')
    children = mp.active_children()
    for child in children:
        try:
            child.terminate()
        except:
            continue
    if logger:
        logger.warning(
            'Incomplete splitting will leave generated files in an incomplete'
            ' state. Please make sure to clean up manually.'
        )
        logger.warning(
            'You may also have to hit CTRL+C again to fully exit, '
            'if the first one does not fully terminate the program.'
        )
        logger.warning('Exiting...')
    sys.exit(1)


def list_files(parsed_args):
    """
    Using the Namespace object generated by argparse, list the files
    that match the given criteria
    """
    if not parsed_args.org:
        parsed_args.logger.error(
            'No valid value was given for option --org. '
            'Please provide one via the CLI or in your config file.'
        )
        sys.exit(1)
    default_edx_buckets = {
        'sql': 'course-data', 'email': 'course-data',
        'log': 'edx-course-data', 'rdx': 'edx-course-data'
    }
    if hasattr(parsed_args, 'aws_cred_file'):
        parsed_args.credentials = cli_utils.find_config(
            parsed_args.aws_cred_file, no_raise=True
        )
    elif parsed_args.credentials is None:
        parsed_args.credentials = cli_utils.find_config(
            os.path.join('~', '.aws', 'credentials'),
            no_raise=True
        )
    client_id = parsed_args.credentials.get(
        parsed_args.profile_name, 'aws_access_key_id',
        fallback=None
    )
    client_secret = parsed_args.credentials.get(
        parsed_args.profile_name, 'aws_secret_access_key',
        fallback=None
    )
    session_token = parsed_args.credentials.get(
        parsed_args.profile_name, 'aws_session_token',
        fallback=None
    )
    start_year = int(parsed_args.begin_date[:4])
    end_year = int(parsed_args.end_date[:4])
    info = aws.BUCKETS.get(parsed_args.file_type)
    # Construct the attribute name for fetching the bucket name from
    # the Namespace object.
    attr = '{t}_bucket'.format(t=parsed_args.file_type)
    bucket_name = getattr(parsed_args, attr, None)
    if not bucket_name:
        bucket_name = default_edx_buckets.get(parsed_args.file_type.lower())
    bucket = aws.make_s3_bucket(
        info['Bucket'].format(bucket=bucket_name), client_id=client_id,
        client_secret=client_secret, session_token=session_token,
        profile_name=parsed_args.profile_name,
    )
    blobs = []
    if parsed_args.latest:
        range_ = range(end_year, start_year - 1, -1)
    else:
        range_ = range(start_year, end_year + 1)
    seen = set()
    prefixes = set()
    for year in range_:
        if parsed_args.latest and seen:
            break
        prefix = info['Prefix'].format(
            site=parsed_args.site or 'edx', year=year,
            date=parsed_args.begin_date, org=parsed_args.org,
            request=parsed_args.request_id or '',
        )
        if prefix in prefixes:
            break
        prefixes.add(prefix)
        blobs = aws.S3Blob.from_prefix(bucket=bucket, prefix=prefix)
        if parsed_args.latest and blobs:
            blobs = sorted(
                blobs, key=lambda b: aws.get_file_date(b.name), reverse=True,
            )
        for blob in blobs:
            if parsed_args.latest and seen:
                break
            if blob.name in seen:
                continue
            fdate = aws.get_file_date(blob.name)
            if parsed_args.begin_date <= fdate <= parsed_args.end_date:
                if parsed_args.json:
                    if parsed_args.names_only:
                        print(blob.name, flush=True)
                    else:
                        print(blob.to_json(), flush=True)
                else:
                    if parsed_args.names_only:
                        print(blob.name, flush=True)
                    else:
                        print(blob, flush=True)
                seen.add(blob.name)
    sys.exit(0 if seen else 1)


def split_log_files(parsed_args):
    """
    Using the Namespace object generated by argparse, parse the given
    tracking log files and put them in the provider destination directory
    """
    files = parsed_args.downloaded_files
    parsed_args.courses = cli_utils.course_listings(parsed_args.courses)
    success = 0
    if len(files) == 1 or parsed_args.dynamic_date:
        for fname in files:
            msg = 'Splitting {f}'.format(f=fname)
            parsed_args.logger.info(msg)
            rc = logs.split_tracking_log(
                filename=fname,
                ddir=parsed_args.destination,
                dynamic_date=parsed_args.dynamic_date,
                courses=parsed_args.courses,
                schema_dir=parsed_args.schema_dir,
            )
            if not rc:
                errmsg = (
                    'No files were extracted while splitting the tracking '
                    'log file {f!r} with the given criteria.'
                )
                parsed_args.logger.warning(errmsg.format(f=fname))
            parsed_args.logger.info('Done splitting {f}'.format(f=fname))
            success += rc
    else:
        success = logs.batch_split_tracking_logs(
            filenames=files,
            ddir=parsed_args.destination,
            dynamic_date=False,
            courses=parsed_args.courses,
            verbose=parsed_args.verbose,
            logger=parsed_args.logger,
            size=parsed_args.jobs,
            schema_dir=parsed_args.schema_dir,
            debug=parsed_args.debug,
        )
    sys.exit(0 if success else 1)


def split_sql_files(parsed_args):
    """
    Split the SQL data archive into separate folders.
    """
    failed = False
    msg = '{w} file name {f}'
    parsed_args.courses = cli_utils.course_listings(parsed_args.courses)
    for fname in parsed_args.downloaded_files:
        parsed_args.logger.info(
            msg.format(f=fname, w='Splitting')
        )
        try:
            to_decrypt = sqls.process_sql_archive(
                archive=fname, ddir=parsed_args.destination,
                include_edge=parsed_args.include_edge,
                courses=parsed_args.courses,
                size=parsed_args.jobs,
                tables_only=parsed_args.tables_only,
                debug=parsed_args.debug,
            )
            if not to_decrypt:
                errmsg = (
                    'No files extracted while splitting the '
                    'contents of {f!r} with the given criteria. '
                    'Moving on'
                )
                parsed_args.logger.warning(errmsg.format(f=fname))
                parsed_args.logger.warning(msg.format(
                    f=fname, w='Done splitting'
                ))
                continue
            parsed_args.logger.info(
                msg.format(f=fname, w='Done splitting')
            )
            if parsed_args.no_decryption:
                continue
            if not parsed_args.tables_only:
                parsed_args.logger.info(
                    msg.format(f=fname, w='Decrypting the contents in')
                )
                try:
                    sqls.batch_decrypt_files(
                        all_files=to_decrypt, size=parsed_args.decryption_batch,
                        verbose=parsed_args.verbose, logger=parsed_args.logger,
                        timeout=parsed_args.decryption_timeout,
                        keepfiles=parsed_args.keep_encrypted,
                        njobs=parsed_args.jobs,
                    )
                except Exception as excp:
                    if parsed_args.fail_fast:
                        raise excp
                    parsed_args.logger.error(excp)
                parsed_args.logger.info(
                    msg.format(f=fname, w='Done decrypting the contents in')
                )
            if parsed_args.unpack_only:
                continue
            orapth = os.path.join('', 'ora', '')
            dirnames = set(
                os.path.dirname(f) for f in to_decrypt if orapth not in f
            )
            parsed_args.logger.info('Making reports from course SQL files')
            if len(dirnames) == 1:
                make_sql_tables = make_sql_tables_seq
            else:
                make_sql_tables = make_sql_tables_par
            try:
                rc = make_sql_tables(
                    dirnames=dirnames, verbose=parsed_args.verbose,
                    logger=parsed_args.logger, fail_fast=parsed_args.fail_fast,
                    debug=parsed_args.debug, schema_dir=parsed_args.schema_dir,
                )
                failed = not rc
            except Exception as excp:
                if parsed_args.fail_fast:
                    raise excp
                parsed_args.logger.error(excp)
                continue
            parsed_args.logger.info(
                'Course reports generated with bundle {f}'.format(f=fname)
            )
        except:
            _, excp, tb = sys.exc_info()
            if isinstance(excp, SystemExit):
                raise excp
            msg = 'Failed to split and decrypt {f}: {e}'
            if parsed_args.debug:
                traces = ['{e}'.format(e=excp)]
                traces += map(str.strip, traceback.format_tb(tb))
                msg = msg.format(f=fname, e='\n'.join(traces))
            else:
                msg = msg.format(f=fname, e=excp)
            parsed_args.logger.error(msg)
            failed = True
    sys.exit(0 if not failed else 1)


def split_email_files(parsed_args):
    """
    Convert a set of email files into a json.gz file to be
    loaded to BigQuery
    """
    if parsed_args.verbose:
        parsed_args.logger.info(
            'Generating json.gz from the email-opt-in CSV file(s)'
        )
    emails.compress_email_files(
        files=parsed_args.downloaded_files,
        ddir=parsed_args.destination,
        schema_dir=parsed_args.schema_dir,
    )
    if parsed_args.verbose:
        parsed_args.logger.info(
            'Done generating json.gz file from the email-opt-in CSV file(s)'
        )


def split_files(parsed_args):
    """
    Split log or SQL files
    """
    if parsed_args.no_courses:
        parsed_args.courses = None
    items = []
    for item in parsed_args.downloaded_files:
        if '*' in item:
            nsubs = 0
            for subitem in glob.iglob(item):
                if not os.path.isfile(subitem):
                    msg = 'Skipping {f!r} because it is not a valid file'
                    parsed_args.logger.warning(msg.format(f=subitem))
                    continue 
                items.append(subitem)
                nsubs += 1
            if nsubs == 0:
                msg = 'The glob pattern {p!r} did not match any file'
                parsed_args.logger.warning(msg.format(p=item))
        else:
            if not os.path.isfile(item):
                msg = 'Skipping {f!r} because it is not a valid file'
                parsed_args.logger.warning(msg.format(f=item))
                continue
            items.append(item)
    parsed_args.downloaded_files = items
    if not parsed_args.downloaded_files:
        parsed_args.logger.error(
            'No valid files were given to simeon split. '
            'Please provide existing files.'
        )
        sys.exit(1)
    if parsed_args.file_type == 'log':
        split_log_files(parsed_args)
    elif parsed_args.file_type == 'sql':
        split_sql_files(parsed_args)
    elif parsed_args.file_type == 'email':
        split_email_files(parsed_args)
    else:
        parsed_args.logger.error(
            'The split command does not support file type {ft}'.format(
                ft=parsed_args.file_type
            )
        )
        sys.exit(1)


def download_files(parsed_args):
    """
    Using the Namespace object generated by argparse, download the files
    that match the given criteria
    """
    default_edx_buckets = {
        'sql': 'course-data', 'email': 'course-data',
        'log': 'edx-course-data', 'rdx': 'edx-course-data'
    }
    if not parsed_args.org:
        parsed_args.logger.error(
            'No valid value was given for option --org. '
            'Please provide one via the CLI or in your config file.'
        )
        sys.exit(1)
    if hasattr(parsed_args, 'aws_cred_file'):
        parsed_args.credentials = cli_utils.find_config(
            parsed_args.aws_cred_file, no_raise=True
        )
    elif parsed_args.credentials is None:
        if parsed_args.verbose:
            parsed_args.logger.info(
                'Looking up S3 credentials'
            )
        parsed_args.credentials = cli_utils.find_config(
            os.path.join('~', '.aws', 'credentials'),
            no_raise=True
        )
    client_id = parsed_args.credentials.get(
        parsed_args.profile_name, 'aws_access_key_id',
        fallback=None
    )
    client_secret = parsed_args.credentials.get(
        parsed_args.profile_name, 'aws_secret_access_key',
        fallback=None
    )
    session_token = parsed_args.credentials.get(
        parsed_args.profile_name, 'aws_session_token',
        fallback=None
    )
    start_year = int(parsed_args.begin_date[:4])
    end_year = int(parsed_args.end_date[:4])
    info = aws.BUCKETS.get(parsed_args.file_type)
    if parsed_args.verbose:
        parsed_args.logger.info(
            'Establishing a connection to S3'
        )
    # Construct the attribute name for fetching the bucket name from
    # the Namespace object.
    attr = '{t}_bucket'.format(t=parsed_args.file_type)
    bucket_name = getattr(parsed_args, attr, None)
    if not bucket_name:
        bucket_name = default_edx_buckets.get(parsed_args.file_type.lower())
    bucket = aws.make_s3_bucket(
        info['Bucket'].format(bucket=bucket_name), client_id=client_id,
        client_secret=client_secret, session_token=session_token,
        profile_name=parsed_args.profile_name,
    )
    if parsed_args.verbose:
        parsed_args.logger.info(
            'Connection to S3 established.'
        )
    blobs = []
    if parsed_args.latest:
        range_ = range(end_year, start_year - 1, -1)
    else:
        range_ = range(start_year, end_year + 1)
    if parsed_args.verbose:
        parsed_args.logger.info(
            'Fetching the blobs matching the given criteria'
        )
    seen = set()
    for year in range_:
        prefix = info['Prefix'].format(
            site=parsed_args.site or 'edx', year=year,
            date=parsed_args.begin_date, org=parsed_args.org,
            request=parsed_args.request_id or '',
        )
        if prefix in seen:
            continue
        seen.add(prefix)
        blobs += aws.S3Blob.from_prefix(bucket=bucket, prefix=prefix)
        if parsed_args.latest and blobs:
            blobs = sorted(
                blobs, key=lambda b: aws.get_file_date(b.name), reverse=True,
            )
            break
    downloads = dict()
    seen = set()
    for blob in blobs:
        if parsed_args.latest and downloads:
            break
        if blob.name in seen:
            continue
        fdate = aws.get_file_date(blob.name)
        if parsed_args.begin_date <= fdate <= parsed_args.end_date:
            fullname = os.path.join(
                parsed_args.destination,
                os.path.basename(os.path.join(*blob.name.split('/')))
            )
            downloads.setdefault(fullname, 0)
            parsed_args.logger.info(
                'Downloading {n} into {f}'.format(n=blob.name, f=fullname)
            )
            blob.download_file(fullname)
            downloads[fullname] += 1
            parsed_args.logger.info(
                'Done downloading {n}'.format(n=blob.name)
            )
            try:
                if parsed_args.file_type != 'sql':
                    parsed_args.logger.info(
                        'Decrypting {f}'.format(f=fullname)
                    )
                if parsed_args.file_type == 'email':
                    parsed_args.downloaded_files = []
                    parsed_args.downloaded_files.append(
                        emails.process_email_file(
                            fname=fullname, verbose=parsed_args.verbose,
                            logger=parsed_args.logger,
                            timeout=parsed_args.decryption_timeout,
                            keepfiles=parsed_args.keep_encrypted
                        )
                    )
                    if parsed_args.verbose:
                        msg = 'Downloaded and decrypted the contents of {f}'
                        parsed_args.logger.info(msg.format(f=fullname))
                elif parsed_args.file_type == 'log':
                    downutils.decrypt_files(
                        fnames=fullname, verbose=parsed_args.verbose,
                        logger=parsed_args.logger,
                        timeout=parsed_args.decryption_timeout,
                    )
                    if parsed_args.verbose:
                        msg = 'Downloaded and decrypted the contents of {f}'
                        parsed_args.logger.info(msg.format(f=fullname))
                downloads[fullname] += 1
                seen.add(blob.name)
            except Exception as excp:
                parsed_args.logger.error(excp)
            cond = all((
                not parsed_args.keep_encrypted,
                parsed_args.file_type != 'sql'
            ))
            if cond:
                if downloads[fullname] == 2:
                    try:
                        os.remove(fullname)
                    except:
                        pass
    if not downloads:
        parsed_args.logger.warning(
            'No files found matching the given criteria'
        )
    if parsed_args.file_type == 'log' and parsed_args.split:
        parsed_args.downloaded_files = []
        for k, v in downloads.items():
            if v == 2:
                k, _ = os.path.splitext(k)
                parsed_args.downloaded_files.append(k)
        if not parsed_args.split_destination:
            parsed_args.destination = os.path.join(
                parsed_args.destination, 'TRACKING_LOGS'
            )
        else:
            parsed_args.destination = parsed_args.split_destination
        split_log_files(parsed_args)
    elif parsed_args.file_type == 'sql' and parsed_args.split:
        parsed_args.downloaded_files = list(downloads)
        if not parsed_args.split_destination:
            parsed_args.destination = os.path.join(
                parsed_args.destination, 'SQL'
            )
        else:
            parsed_args.destination = parsed_args.split_destination
        split_sql_files(parsed_args)
    elif parsed_args.file_type == 'email' and parsed_args.split:
        if not parsed_args.split_destination:
            parsed_args.destination = os.path.join(
                parsed_args.destination, 'email_opt_in'
            )
        else:
            parsed_args.destination = parsed_args.split_destination
        split_email_files(parsed_args)
    failed = not downloads or any(v != 2 for v in downloads.values())
    sys.exit(1 if failed else 0)


def push_to_bq(parsed_args):
    """
    Push to BigQuery
    """
    if not parsed_args.project:
        parsed_args.logger.error(
            'No GCP project given in the command line. '
            'None was found in config file(s) either. '
            'Aborting...'
        )
        sys.exit(1)
    if not parsed_args.items:
        parsed_args.logger.warning('No items to process. Exiting...')
        sys.exit(1)
    parsed_args.logger.info('Connecting to BigQuery')
    try:
        if parsed_args.service_account_file is not None:
            client = gcp.BigqueryClient.from_service_account_json(
                parsed_args.service_account_file,
                project=parsed_args.project
            )
        else:
            client = gcp.BigqueryClient(
                project=parsed_args.project
            )
    except Exception as excp:
        errmsg = 'Failed to connect to BigQuery: {e}'
        parsed_args.logger.error(errmsg.format(e=excp))
        parsed_args.logger.error(
            'The error may be from an invalid service account file'
        )
        sys.exit(1)
    all_jobs = []
    storage = parsed_args.use_storage
    tmp = parsed_args.items
    parsed_args.items = []
    for item in tmp:
        if item.startswith('gs'):
            parsed_args.items.append(item)
        elif '*' in item:
            parsed_args.items.extend(glob.iglob(item))
        else:
            parsed_args.items.append(item)
    for item in parsed_args.items:
        parsed_args.use_storage = storage or item.startswith('gs://')
        if not parsed_args.use_storage and not os.path.exists(item):
            errmsg = 'Skipping {f!r}. It does not exist.'
            parsed_args.logger.warning(errmsg.format(f=item))
            if parsed_args.fail_fast:
                parsed_args.logger.error('Exiting...')
                sys.exit(1)
            continue
        if os.path.isdir(item):
            loader = client.load_tables_from_dir
            appender = all_jobs.extend
        else:
            loader = client.load_one_file_to_table
            appender = all_jobs.append
        parsed_args.logger.info(
            'Loading item(s) in {f!r} to BigQuery'.format(f=item)
        )
        # Use .job_id here, so we don't lug LoadJob objects in memory.
        # We would still need to make network calls
        # to check on their statuses when wait_for_loads is True.
        try:
            job = loader(
                item, file_type=parsed_args.file_type,
                project=parsed_args.project, create=parsed_args.create,
                append=parsed_args.append, use_storage=parsed_args.use_storage,
                bucket=parsed_args.bucket,
                max_bad_rows=parsed_args.max_bad_rows,
                schema_dir=parsed_args.schema_dir,
                format_=parsed_args.file_format,
                patch=parsed_args.update_description,
            )
            if parsed_args.wait_for_loads:
                if isinstance(job, (list, tuple)):
                    appender(map(lambda j: j.job_id, job))
                else:
                    appender(job.job_id)
            parsed_args.logger.info(
                'Created BigQuery load job(s) for item(s) in {f!r}'.format(f=item)
            )
        except Exception as excp:
            msg = (
                'Failed to create BigQuery load job(s) for item(s) '
                'in {f!r}: {e}'
            )
            parsed_args.logger.error(msg.format(f=item, e=excp))
    if not all_jobs and parsed_args.wait_for_loads:
        errmsg = (
            'No items processed. Perhaps, the given directory is empty?'
        )
        parsed_args.logger.error(errmsg)
        sys.exit(1)
    if not parsed_args.wait_for_loads:
        msg = (
            '{c} BigQuery data load job(s) started. Please consult your '
            'BigQuery console for more details about the status of said jobs.'
        )
        parsed_args.logger.info(msg.format(c=len(parsed_args.items)))
        sys.exit(0)
    parsed_args.logger.info(
        'Checking on the status of the submitted load job(s)...'
    )
    all_jobs = wait_for_bq_job_ids(all_jobs, client)
    err_count = 0
    for job, errors in all_jobs.items():
        if errors:
            for err in client.extract_error_messages(errors):
                parsed_args.logger.error(err)
            err_count += 1
    if err_count:
        msg = (
            'Out of {j} load job(s) submitted, {f} failed to '
            'complete successfully'
        )
        parsed_args.logger.error(msg.format(
            j=len(all_jobs), f=err_count
        ))
        sys.exit(1)
    parsed_args.logger.info(
        '{c} item(s) loaded to BigQuery'.format(c=len(all_jobs))
    )


def push_to_gcs(parsed_args):
    """
    Push to Storage
    """
    if not parsed_args.bucket:
        parsed_args.logger.error(
            'No valid GCP bucket given in the command line. '
            'None was found in config file(s) either. '
            'Aborting...'
        )
        sys.exit(1)
    if not parsed_args.items:
        parsed_args.logger.warning('No items to process. Exiting...')
        sys.exit(1)
    parsed_args.logger.info(
        'Connecting to Google Cloud Storage'
    )
    try:
        if parsed_args.service_account_file is not None:
            client = gcp.GCSClient.from_service_account_json(
                parsed_args.service_account_file,
                project=parsed_args.project
            )
        else:
            client = gcp.GCSClient(
                project=parsed_args.project
            )
    except Exception as excp:
        errmsg = 'Failed to connect to Google Cloud Storage: {e}'
        parsed_args.logger.error(errmsg.format(e=excp))
        parsed_args.logger.error(
            'The error may be from an invalid service account file'
        )
        sys.exit(1)
    failed = False
    nitems = 0
    tmp = parsed_args.items
    parsed_args.items = []
    for item in tmp:
        if '*' in item:
            parsed_args.items.extend(glob.iglob(item))
        else:
            parsed_args.items.append(item)
    for item in parsed_args.items:
        if not os.path.exists(item):
            errmsg = 'Skipping {f!r}. It does not exist.'
            parsed_args.logger.warning(errmsg.format(f=item))
            if parsed_args.fail_fast:
                parsed_args.logger.error('Error encountered. Exiting...')
                sys.exit(1)
            continue
        if os.path.isdir(item):
            loader = client.load_dir
        else:
            loader = client.load_one_file_to_gcs
        try:
            parsed_args.logger.info(
                'Loading {f} to GCS'.format(f=item)
            )
            loader(
                item, parsed_args.file_type,
                parsed_args.bucket,
            )
            parsed_args.logger.info(
                'Done loading {f} to GCS'.format(f=item)
            )
            nitems += 1
        except Exception as excp:
            errmsg = 'Failed to load {f} to GCS: {e}'
            parsed_args.logger.error(errmsg.format(f=item, e=excp))
            parsed_args.logger.error(
                'The error may be from an invalid service account file'
            )
            if parsed_args.fail_fast:
                parsed_args.logger.error('Exiting...')
                sys.exit(1)
            failed = True
    sys.exit(1 if failed or nitems == 0 else 0)


def _clean_up_extras(parsed_args):
    """
    Clean the parsed_args.extras dictionary and remove anything
    that's passed via CLI or is a defined parameters to functions
    making secondary datasets.
    """
    if not isinstance(parsed_args.extra_args, dict):
        return
    keys = list(parsed_args.extra_args)
    params = ['course_id', 'client', 'wait']
    for k in keys:
        if hasattr(parsed_args, k):
            parsed_args.extra_args.pop(k, None)
    for k in params:
        parsed_args.extra_args.pop(k, None)


def push_generated_files(parsed_args):
    """
    Using the Namespace object generated by argparse, push data files
    to a target destination
    """
    if parsed_args.in_files:
        try:
            parsed_args.items = cli_utils.items_from_files(parsed_args.items)
        except Exception as excp:
            parsed_args.logger.error(
                'The given items are not valid text files: {e}'.format(e=excp)
            )
            sys.exit(1)
    # If file type in log or sql, filter by course ID
    if parsed_args.file_type in ('log', 'sql',) and not parsed_args.no_courses:
        parsed_args.items = cli_utils.filter_generated_items(
            parsed_args.items, parsed_args.courses
        )
    try:
        # Not sure if this is going to reduce memory usage by this process,
        # but worth the try
        delattr(parsed_args, 'courses')
    except AttributeError:
        pass
    if parsed_args.destination == 'bq':
        push_to_bq(parsed_args)
    else:
        push_to_gcs(parsed_args)


def make_secondary_tables(parsed_args):
    """
    Generate secondary datasets that rely on existing datasets
    and tables.
    """
    if not parsed_args.project:
        parsed_args.logger.error(
            'No GCP project given in the command line. '
            'None was found in config file(s) either. '
            'Aborting...'
        )
        sys.exit(1)
    if not parsed_args.course_ids:
        parsed_args.logger.info('No items to process')
        sys.exit(0)
    parsed_args.extra_args = cli_utils.process_extra_args(
        parsed_args.extra_args
    )
    parsed_args.logger.info('Connecting to BigQuery')
    try:
        if parsed_args.service_account_file is not None:
            client = gcp.BigqueryClient.from_service_account_json(
                parsed_args.service_account_file,
                project=parsed_args.project
            )
        else:
            client = gcp.BigqueryClient(
                project=parsed_args.project
            )
    except Exception as excp:
        errmsg = (
            'Failed to connect to BigQuery: {e}\n'
            'The error may be from an invalid service account file'
        )
        parsed_args.logger.error(errmsg.format(e=excp))
        sys.exit(1)
    parsed_args.logger.info('Connection established')
    if parsed_args.in_files:
        parsed_args.course_ids = cli_utils.items_from_files(
            parsed_args.course_ids
        )
    all_jobs = dict()
    _clean_up_extras(parsed_args)
    if len(parsed_args.course_ids) == 1:
        course_id = list(parsed_args.course_ids)[0]
        parsed_args.logger.info(
            'Making secondary tables for course ID {cid}'.format(
                cid=course_id
            )
        )
        all_jobs[course_id] = make_tables_from_sql(
            tables=parsed_args.tables, course_id=course_id, client=client,
            project=parsed_args.project, append=parsed_args.append,
            geo_table=parsed_args.geo_table,
            youtube_table=parsed_args.youtube_table,
            wait=parsed_args.wait_for_loads, fail_fast=parsed_args.fail_fast,
            query_dir=parsed_args.query_dir, schema_dir=parsed_args.schema_dir,
            **parsed_args.extra_args
        )
    else:
        all_jobs.update(make_tables_from_sql_par(
            tables=parsed_args.tables, courses=parsed_args.course_ids,
            safile=parsed_args.service_account_file, 
            project=parsed_args.project,
            append=parsed_args.append, geo_table=parsed_args.geo_table,
            youtube_table=parsed_args.youtube_table,
            wait=parsed_args.wait_for_loads, size=parsed_args.jobs,
            logger=parsed_args.logger, fail_fast=parsed_args.fail_fast,
            query_dir=parsed_args.query_dir, schema_dir=parsed_args.schema_dir,
            **parsed_args.extra_args
        ))
    errors = 0
    num_queries = 0
    parsed_args.logger.info('Checking for errors...')
    for course_id, job_errors in all_jobs.items():
        for table, error_dict in job_errors.items():
            num_queries += 1
            if error_dict:
                msg = 'Making {t} for {c} failed with the following: {e}'
                parsed_args.logger.error(msg.format(
                    t=table, c=course_id,
                    e='\n'.join(client.extract_error_messages(error_dict))
                ))
                errors += 1
    if errors:
        sys.exit(1)
    if parsed_args.wait_for_loads:
        if num_queries > 1:
            msg = (
                '{c} queries completed and destination '
                'tables have been refreshed.'
            )
        else:
            msg = 'The corresponding ran successfully.'
        parsed_args.logger.info(msg.format(c=num_queries))
        sys.exit(0)
    msg = (
        '{c} BigQuery query jobs started. Please consult your '
        'BigQuery console for more details about the status of said jobs.'
    )
    parsed_args.logger.info(msg.format(c=num_queries))


def main():
    """
    Entry point
    """
    global logger
    COMMANDS = {
        'list': list_files,
        'download': download_files,
        'split': split_files,
        'push': push_generated_files,
        'report': make_secondary_tables,
    }
    parser = cli_utils.CustomArgParser(
        description=__doc__,
        formatter_class=RawDescriptionHelpFormatter,
        allow_abbrev=False,
        epilog=cli_utils.CLI_MAIN_EPILOG,
        prog='simeon',
    )
    parser.add_argument(
        '--quiet', '-Q',
        help='Only print error messages to standard streams.',
        action='store_false',
        dest='verbose',
    )
    parser.add_argument(
        '--debug', '-B',
        help='Show some stacktrace if simeon stops because of a fatal error',
        action='store_true',
    )
    parser.add_argument(
        '--config-file', '-C',
        help=(
            'The INI configuration file to use for default arguments.'
        ),
    )
    parser.add_argument(
        '--log-file', '-L',
        help='Log file to use when simeon prints messages. Default: stdout',
        type=FileType('a'),
        default=sys.stdout,
    )
    parser.add_argument(
        '--version', '-v',
        action='version',
        version='%(prog)s {v}'.format(v=simeon.__version__)
    )
    subparsers = parser.add_subparsers(
        description='Choose a subcommand to carry out a task with simeon',
        dest='command',
        metavar='subcommand',
    )
    subparsers.required = True
    downloader = subparsers.add_parser(
        'download',
        help='Download edX research data with the given criteria',
        description=(
            'Download edX research data with the given criteria below'
        ),
        allow_abbrev=False,
    )
    downloader.set_defaults(command='download')
    downloader.add_argument(
        '--file-type', '-f',
        help='The type of files to get. Default: %(default)s',
        choices=['email', 'sql', 'log', 'rdx'],
        required=True,
    )
    downloader.add_argument(
        '--email-bucket',
        help='The S3 bucket name for email files. Default: course-data',
    )
    downloader.add_argument(
        '--sql-bucket',
        help='The S3 bucket name for SQL files. Default: course-data',
    )
    downloader.add_argument(
        '--log-bucket',
        help='The S3 bucker name for tracking log files. Default: edx-course-data',
    )
    downloader.add_argument(
        '--rdx-bucket',
        help='The S3 bucket name for the RDX files. Default: edx-course-data',
    )
    downloader.add_argument(
        '--destination', '-d',
        help='Directory where to download the file(s). Default: %(default)s',
        default=os.getcwd(),
    )
    downloader.add_argument(
        '--begin-date', '-b',
        help=(
            'Start date of the download timeframe. '
            'Default: %(default)s'
        ),
        default=aws.BEGIN_DATE,
        type=cli_utils.parsed_date
    )
    downloader.add_argument(
        '--end-date', '-e',
        help=(
            'End date of the download timeframe. '
            'Default: %(default)s'
        ),
        default=aws.END_DATE,
        type=cli_utils.parsed_date
    )
    downloader.add_argument(
        '--latest', '-L',
        help='Download the latest file only',
        action='store_true',
    )
    downloader.add_argument(
        '--org', '-o',
        help='The organization whose data is fetched. Default: %(default)s',
    )
    downloader.add_argument(
        '--site', '-s',
        help='The edX site from which to pull data. Default: %(default)s',
        choices=['edge', 'edx', 'patches'],
    )
    downloader.add_argument(
        '--aws-cred-file', '-a',
        help='The AWS INI credentials file to use to connect to S3.',
        type=cli_utils.find_config,
        dest='credentials',
    )
    downloader.add_argument(
        '--profile-name', '-p',
        help=(
            'The section name in the credentials file containing the '
            'acess and secret keys needed to access edX\'s S3 bucket. '
        ),
    )
    downloader.add_argument(
        '--split', '-S',
        help='Split downloaded SQL bundles or tracking logs',
        action='store_true',
    )
    downloader.add_argument(
        '--split-destination', '-D',
        help=(
            'The directory in which to put the split files if --split is '
            'given with this subcommand'
        ),
    )
    downloader.add_argument(
        '--jobs', '-j',
        help=(
            'Number of processes/threads to use when splitting multiple '
            'files using multi threading or processing. Default: %(default)s'
        ),
        default=mp.cpu_count(),
        type=int,
    )
    downloader.add_argument(
        '--schema-dir', '-R',
        help=(
            'Directory where to find schema files. '
            'Default: {d}'.format(d=SCHEMA_DIR)
        ),
    )
    downloader.add_argument(
        '--dynamic-date', '-m',
        help=(
            'If splitting the downloaded files, use the '
            'dates from the records to make tracking log file names. '
            'Otherwise, the dates in the GZIP file names are used.'
        ),
        action='store_true',
    )
    downloader.add_argument(
        '--request-id', '-r',
        help='Request ID when downloading RDX files',
    )
    downloader.add_argument(
        '--decryption-timeout', '-t',
        help='Number of seconds to wait for the decryption of files.',
        type=int,
    )
    downloader.add_argument(
        '--unpack-only', '-u',
        help=(
            'If --split is used, only unpack the SQL archive and '
            'decrypt the encrypted files. '
            'Don\'t make any of the json.gz files for course_axis, '
            'studentmodule, etc.'
        ),
        action='store_true',
    )
    downloader.add_argument(
        '--decryption-batch', '-g',
        help=(
            'Number of files to batch decrypt using gpg. '
            'Expected values are between 1 and 50. '
            'Default: %(default)s'
        ),
        type=cli_utils.NumberRange(int),
        default=25,
    )
    cdgroup = downloader.add_mutually_exclusive_group(required=False)
    cdgroup.add_argument(
        '--no-courses', '-V',
        help=(
            'DO NOT try to limit the generated files to only some course IDs. '
            'Therefore, simeon will split everything.'
        ),
        action='store_true',
    )
    cdgroup.add_argument(
        '--courses', '-c',
        help=(
            'A list of white space separated course IDs whose data files '
            'are unpacked and decrypted.'
        ),
        nargs='*',
    )
    cdgroup.add_argument(
        '--clistings-file', '-l',
        help=(
            'Path to a file with one course ID per line. The file is expected'
            ' to have no header row.'
        ),
        type=cli_utils.courses_from_file,
    )
    downloader.add_argument(
        '--no-decryption', '-N',
        help='Don\'t decrypt the unpacked SQL files.',
        action='store_true',
    )
    downloader.add_argument(
        '--include-edge', '-E',
        help='Include the edge site files when splitting SQL data packages.',
        action='store_true',
    )
    downloader.add_argument(
        '--keep-encrypted', '-k',
        help='Keep the encrypted files after decrypting them',
        action='store_true',
    )
    downloader.add_argument(
        '--tables-only', '-T',
        help=(
            'Don\'t do any unpacking of the SQL archive. Use the latter to '
            'get directories that already contain SQL files to use to make '
            'files to load to BigQuery.'
        ),
        action='store_true',
    )
    downloader.add_argument(
        '--fail-fast', '-F',
        help=(
            'Force simeon to stop splitting if any error is encountered. '
            'Otherwise, simeon reports the error and moves on to the next thing.'
        ),
        action='store_true',
    )
    lister = subparsers.add_parser(
        'list',
        help='List edX research data with the given criteria',
        description=(
            'List edX research data with the given criteria below'
        ),
        allow_abbrev=False,
    )
    lister.set_defaults(command='list')
    lister.add_argument(
        '--file-type', '-f',
        help='The type of files to list. Default: %(default)s',
        choices=['email', 'sql', 'log', 'rdx'],
        required=True,
    )
    lister.add_argument(
        '--email-bucket',
        help='The S3 bucket name for email files. Default: course-data',
    )
    lister.add_argument(
        '--sql-bucket',
        help='The S3 bucket name for SQL files. Default: course-data',
    )
    lister.add_argument(
        '--log-bucket',
        help='The S3 bucker name for tracking log files. Default: edx-course-data',
    )
    lister.add_argument(
        '--rdx-bucket',
        help='The S3 bucket name for the RDX files. Default: edx-course-data',
    )
    lister.add_argument(
        '--begin-date', '-b',
        help=(
            'Start date of the listing timeframe. '
            'Default: %(default)s'
        ),
        default=aws.BEGIN_DATE,
        type=cli_utils.parsed_date
    )
    lister.add_argument(
        '--end-date', '-e',
        help=(
            'End date of the listing timeframe. '
            'Default: %(default)s'
        ),
        default=aws.END_DATE,
        type=cli_utils.parsed_date
    )
    lister.add_argument(
        '--latest', '-L',
        help='List only the latest file for the given file type',
        action='store_true',
    )
    lister.add_argument(
        '--names-only', '-n',
        help='Only print the names of the matched blobs',
        action='store_true',
    )
    lister.add_argument(
        '--org', '-o',
        help='The organization whose data is listed. Default: %(default)s',
    )
    lister.add_argument(
        '--site', '-s',
        help='The edX site from which to list data. Default: %(default)s',
        choices=['edge', 'edx', 'patches'],
    )
    lister.add_argument(
        '--aws-cred-file', '-a',
        help='The AWS INI credentials file to use to connect to S3.',
        type=cli_utils.find_config,
        dest='credentials',
    )
    lister.add_argument(
        '--profile-name', '-p',
        help=(
            'The section name in the credentials file containing the '
            'acess and secret keys needed to access edX\'s S3 bucket.'
        ),
    )
    lister.add_argument(
        '--request-id', '-r',
        help='Request ID when listing RDX files',
    )
    lister.add_argument(
        '--json', '-j',
        help='Format the file listing in JSON',
        action='store_true',
    )
    splitter = subparsers.add_parser(
        'split',
        help='Split downloaded tracking log or SQL files',
        description='Split downloaded tracking log or SQL files',
        allow_abbrev=False,
    )
    splitter.set_defaults(command='split')
    splitter.add_argument(
        'downloaded_files',
        help='List of tracking log or SQL archives to split',
        nargs='+',
    )
    splitter.add_argument(
        '--file-type', '-f',
        help='The file type of the items provided. Default: %(default)s',
        required=True,
        choices=['log', 'sql', 'email'],
    )
    splitter.add_argument(
        '--schema-dir', '-R',
        help=(
            'Directory where to find schema files. '
            'Default: {d}'.format(d=SCHEMA_DIR)
        ),
    )
    splitter.add_argument(
        '--no-decryption', '-N',
        help='Don\'t decrypt the unpacked SQL files.',
        action='store_true',
    )
    splitter.add_argument(
        '--jobs', '-j',
        help=(
            'Number of processes/threads to use when processing multiple '
            'files using multi threading or processing. Default: %(default)s'
        ),
        default=mp.cpu_count(),
        type=int,
    )
    splitter.add_argument(
        '--include-edge', '-E',
        help='Include the edge site files when splitting SQL data packages.',
        action='store_true',
    )
    splitter.add_argument(
        '--keep-encrypted', '-k',
        help='Keep the encrypted files after decrypting them',
        action='store_true',
    )
    splitter.add_argument(
        '--decryption-timeout', '-t',
        help='Number of seconds to wait for the decryption of files.',
        type=int,
    )
    splitter.add_argument(
        '--destination', '-d',
        help=(
            'Directory where to place the files from splitting the item(s).'
            ' Default: %(default)s'
        ),
        default=os.getcwd(),
    )
    csgroup = splitter.add_mutually_exclusive_group(required=False)
    csgroup.add_argument(
        '--no-courses', '-V',
        help=(
            'DO NOT try to limit the generated files to only some course IDs. '
            'Therefore, simeon will split everything.'
        ),
        action='store_true',
    )
    csgroup.add_argument(
        '--courses', '-c',
        help=(
            'A list of white space separated course IDs whose data files '
            'are unpacked and decrypted.'
        ),
        nargs='*',
    )
    csgroup.add_argument(
        '--clistings-file', '-l',
        help=(
            'Path to a file with one course ID per line. The file is expected'
            ' to have no header row.'
        ),
        type=cli_utils.courses_from_file,
    )
    splitter.add_argument(
        '--dynamic-date', '-m',
        help=(
            'Use the dates from the records to make tracking log file names. '
            'Otherwise, the dates in the GZIP file names are used.'
        ),
        action='store_true',
    )
    splitter.add_argument(
        '--tables-only', '-T',
        help=(
            'Don\'t do any unpacking of the SQL archive. Use the latter to '
            'get directories that already contain SQL files to use to make '
            'files to load to BigQuery.'
        ),
        action='store_true',
    )
    splitter.add_argument(
        '--unpack-only', '-u',
        help=(
            'Only unpack the archive and decrypt the encrypted files. '
            'Don\'t make any of the json.gz files for course_axis, '
            'studentmodule, etc.'
        ),
        action='store_true',
    )
    splitter.add_argument(
        '--fail-fast', '-F',
        help=(
            'Force simeon to stop splitting if any error is encountered. '
            'Otherwise, simeon reports the error and moves on to the next thing.'
        ),
        action='store_true',
    )
    splitter.add_argument(
        '--decryption-batch', '-g',
        help=(
            'Number of files to batch decrypt using gpg. '
            'Expected values are between 1 and 50. '
            'Default: %(default)s'
        ),
        type=cli_utils.NumberRange(int),
        default=25,
    )
    pusher = subparsers.add_parser(
        'push',
        help='Push the generated data files to some target destination',
        description=(
            'Push to the given items to Google Cloud Storage or BigQuery'
        ),
        allow_abbrev=False,
    )
    pusher.set_defaults(command='push')
    pusher.add_argument(
        'destination',
        help='Sink for the generated data files',
        choices=['gcs', 'bq']
    )
    pusher.add_argument(
        'items',
        help=(
            'The items (files or folders) to push to GCS or BigQuery, or '
            'paths to txt files containing the actual paths of items to be '
            'loaded (if --in-files is provided)'
        ),
        nargs='+',
    )
    pusher.add_argument(
        '--project', '-p',
        help='GCP project associated with the target sink',
    )
    pusher.add_argument(
        '--bucket', '-b',
        help='GCS bucket name associated with the target sink',
        type=cli_utils.gcs_bucket,
    )
    pusher.add_argument(
        '--service-account-file', '-S',
        help='The service account to carry out the data load',
        type=cli_utils.optional_file
    )
    pusher.add_argument(
        '--max-bad-rows', '-m',
        help=(
            'Max number of bad rows to allow when loading data to BigQuery. '
            'Default: %(default)s'
        ),
        type=int,
        default=0,
    )
    pusher.add_argument(
        '--file-type', '-f',
        help='The type of files to push. Default: %(default)s',
        choices=['email', 'sql', 'log', 'rdx', 'cold'],
        required=True,
    )
    pusher.add_argument(
        '--file-format', '-M',
        help='The format of the file to load. Default: %(default)s',
        choices=['json', 'csv'],
        default='json',
    )
    pusher.add_argument(
        '--no-create', '-n',
        help=(
            'Don\'t create destination tables and datasets. '
            'They must already exist for the push operation '
            'to work.'
        ),
        action='store_false',
        dest='create',
    )
    pusher.add_argument(
        '--append', '-a',
        help=(
            'Whether to append to destination tables if they exist'
            ' when pushing data to BigQuery'
        ),
        action='store_true',
    )
    pusher.add_argument(
        '--schema-dir', '-R',
        help=(
            'Directory where to find schema files. '
            'Default: {d}'.format(d=SCHEMA_DIR)
        ),
    )
    pusher.add_argument(
        '--update-description', '-u',
        help=(
            'Applies only to push bq and updates the description '
            'of the destination table with the "description" value'
            ' from the corresponding schema file'
        ),
        action='store_true',
    )
    pusher.add_argument(
        '--use-storage', '-s',
        help='Whether to use GCS for actual files when loading to bq',
        action='store_true',
    )
    pusher.add_argument(
        '--fail-fast', '-F',
        help=(
            'Force simeon to quit as soon as an error is encountered'
            ' with any of the given items.'
        ),
        action='store_true',
    )
    pusher.add_argument(
        '--wait-for-loads', '-w',
        help=(
            'Wait for asynchronous BigQuery load jobs to finish. '
            'Otherwise, simeon creates load jobs and exits.'
        ),
        action='store_true',
    )
    pusher.add_argument(
        '--in-files', '-i',
        help=(
            'Whether the provided items are text files that contain the '
            'paths to be loaded to BigQuery. This option only applies to bq.'
        ),
        action='store_true',
    )
    cpgroup = pusher.add_mutually_exclusive_group(required=False)
    cpgroup.add_argument(
        '--no-courses', '-V',
        help=(
            'DO NOT try to limit the loaded files to only some course IDs. '
            'Therefore, simeon will push everything.'
        ),
        action='store_true',
    )
    cpgroup.add_argument(
        '--courses', '-c',
        help=(
            'A list of white space separated course IDs whose data files '
            'are pushed to the target destination.'
        ),
        nargs='*',
    )
    cpgroup.add_argument(
        '--clistings-file', '-l',
        help=(
            'Path to a file with one course ID per line. The file is expected'
            ' to have no header row. '
            'Only files whose names match the course IDs are pushed.'
        ),
        type=cli_utils.course_paths_from_file,
    )
    reporter = subparsers.add_parser(
        'report',
        help='Make course reports using the datasets and tables in BigQuery',
        description=(
            'Make course reports using the datasets and tables in BigQuery'
        ),
        allow_abbrev=False,
    )
    reporter.set_defaults(command='report')
    reporter.add_argument(
        'course_ids',
        help=(
            'Course IDs whose secondary datasets are generated, '
            'or paths to txt files containing course IDs '
            '(if --in-files is provided)'
        ),
        nargs='+',
    )
    reporter.add_argument(
        '--project', '-p',
        help='GCP project associated with the tables to query',
    )
    reporter.add_argument(
        '--service-account-file', '-S',
        help='The service account to carry out the data load',
        type=cli_utils.optional_file
    )
    reporter.add_argument(
        '--query-dir', '-q',
        help=(
            'The directory where SQL query files live. '
            'Default: {d}'.format(d=QUERY_DIR)
        ),
    )
    reporter.add_argument(
        '--schema-dir', '-R',
        help=(
            'Directory where to find schema files. '
            'Default: {d}'.format(d=SCHEMA_DIR)
        ),
    )
    reporter.add_argument(
        '--append', '-a',
        help=(
            'Whether to append to destination tables if they exist'
        ),
        action='store_true',
    )
    reporter.add_argument(
        '--no-reorder', '-O',
        help=(
            'DO NOT reorder the given table names. Generate them in the '
            'order provided.'
        ),
        action='store_true',
    )
    reporter.add_argument(
        '--tables', '-t',
        help=(
            'Table or tables to be computed using corresponding query files. '
            'Default: {t}'.format(t=' '.join(cli_utils.REPORT_TABLES))
        ),
        nargs='*',
        default=cli_utils.REPORT_TABLES,
        action=cli_utils.TableOrderAction,
    )
    reporter.add_argument(
        '--geo-table', '-g',
        help=(
            'The fully qualified name of the geolocation table '
            'to join to modal_ip to extract geolocation information '
            'for IP addresses.'
        ),
        type=cli_utils.bq_table,
    )
    reporter.add_argument(
        '--youtube-table', '-y',
        help=(
            'Fully qualified name of the table with YouTube video metadata'
            ' to use when making the video_axis tables.'
        ),
        type=cli_utils.bq_table,
    )
    reporter.add_argument(
        '--fail-fast', '-F',
        help=(
            'Force simeon to quit as soon as an error is encountered'
            ' with any of the given items.'
        ),
        action='store_true',
    )
    reporter.add_argument(
        '--wait-for-loads', '-w',
        help=(
            'Wait for asynchronous BigQuery query jobs to finish. '
            'Otherwise, simeon creates query jobs and exits.'
        ),
        action='store_true',
    )
    reporter.add_argument(
        '--in-files', '-i',
        help=(
            'Whether the provided course ID arguments are text '
            'files that contain the course IDs. One ID per line'
        ),
        action='store_true',
    )
    reporter.add_argument(
        '--jobs', '-j',
        help=(
            'Number of processes/threads to use when processing multiple '
            'course IDs using multi threading or processing. '
            'Default: %(default)s'
        ),
        default=mp.cpu_count(),
        type=int,
    )
    reporter.add_argument(
        '--extra-args', '-x',
        help=(
            'Extra arguments to use as placeholder values for variables '
            'specified in query files. Expected format is: variable1=value1,'
            'variable2=value2,...,variablen=valuen'
        ),
    )
    args = parser.parse_args()
    args.logger = cli_utils.make_logger(
        verbose=args.verbose,
        stream=args.log_file,
        user='SIMEON:{cmd}'.format(cmd=args.command.upper()),
    )
    # Get simeon configurations and plug them in wherever
    # a CLI option is not given
    try:
        configs = cli_utils.find_config(args.config_file)
    except Exception as excp:
        args.logger.error(str(excp).replace('\n', ' '))
        sys.exit(1)
    for k, v in cli_utils.CONFIGS.items():
        for (attr, cgetter) in v:
            cli_arg = getattr(args, attr, None)
            config_arg = cgetter(configs, k, attr, fallback=None)
            if attr.replace('-', '_') == 'clistings_file':
                try:
                    if args.command == 'push':
                        config_arg = cli_utils.course_paths_from_file(
                            config_arg
                        )
                    elif args.command not in ('list',):
                        config_arg = cli_utils.courses_from_file(config_arg)
                except Exception as excp:
                    args.logger.error(str(excp).replace('\n', ' '))
                    sys.exit(1)
            if not cli_arg and config_arg:
                setattr(args, attr, config_arg)
    # Combine --courses with --clistings-file
    if hasattr(args, 'courses') and hasattr(args, 'clistings_file'):
        if not getattr(args, 'courses', None):
            args.courses = args.clistings_file
    # Signal handling for the usual interrupters
    # List shortened because Windows is too annoying
    # to know of many of the Unix interrupting signals.
    # Also, set the global logger variable, so the signal handler can use it.
    sigs = [signal.SIGABRT, signal.SIGTERM, signal.SIGINT]
    logger = args.logger
    if cli_utils.is_parallel(args):
        for sig in sigs:
            signal.signal(sig, bail_out)
    # Call the function matching the given command
    try:
        COMMANDS.get(args.command)(args)
    except KeyboardInterrupt:
        args.logger.error('Interrupted by the user')
        sys.exit(1)
    except BrokenPipeError:
        # This should be raised by list_files.
        # Redirect sys.stdout and sys.stderr streams to devnull
        # Return exit code 0, since the cause of this exception
        # may be because some program from the shell decided to
        # close its end of the shell PIPE. Therefore, that's not
        # an issue to simeon
        devnull = os.open(os.devnull, os.O_WRONLY)
        os.dup2(devnull, sys.stdout.fileno())
        os.dup2(devnull, sys.stderr.fileno())
        sys.exit(0)
    except:
        _, excp, tb = sys.exc_info()
        if isinstance(excp, (EarlyExitError, SystemExit)):
            raise excp
        msg = str(excp)
        if args.debug:
            traces = [msg]
            traces += map(str.strip, traceback.format_tb(tb))
            msg = msg.format(c=args.command, e='\n'.join(traces))
        args.logger.error(msg)
        sys.exit(1)


if __name__ == '__main__':
    main()



"""
Entry point for the cli if invoked with python -m simeon
"""
import os
import sys


# Remove '' and current working directory from the first entry
# of sys.path, if present to avoid using current directory
# when invoked as python -m pip <command>
if sys.path[0] in ("", os.getcwd()):
    sys.path.pop(0)

if __name__ == '__main__':
    from simeon.scripts.simeon import main as _main
    sys.exit(_main())



"""
Integration tests for the simeon CLI tool.
These can be quite expensive since they read from and write to files.
"""
import os
import shlex
import subprocess as sb
import unittest


def _run_command(cmd):
    """
    Run the given command and return the generated Popen object
    """
    proc = sb.Popen(shlex.split(cmd), stderr=sb.PIPE, stdout=sb.PIPE)
    proc.wait()
    return proc


class SimeonCLIIntegration(unittest.TestCase):
    """
    Test the simeon CLI command
    """
    def setUp(self):
        self.this_dir = os.path.dirname(os.path.abspath(__file__))
        self.log_file = os.path.join(
            self.this_dir, 'fixtures', 'file.log'
        )
        self.config_file = os.path.join(
            self.this_dir, 'fixture', 'config.ini'
        )
        self.good_subcommands = [
            'download', 'list', 'push', 'split'
        ]
        self.bad_subcommands = [
            'upload', 'inspect', 'run', 'stupid', 'spli', 'donwland',
            'psh', 'splt', 'lst', 'upoad', 'uload'
        ]
        self.global_options = [
            '--log-file {f}'.format(f=self.log_file),
            '--config-file {f}'.format(f=self.config_file),
            '--quiet', '--debug'
        ]
        self.proc = None

    def tearDown(self):
        if self.proc:
            if self.proc.stdout:
                self.proc.stdout.close()
            if self.proc.stderr:
                self.proc.stderr.close()
        for file_ in (self.log_file, self.config_file):
            try:
                os.remove(file_)
            except OSError:
                continue

    def test_good_subcommands(self):
        """
        Test that the subcommands have not been changed.
        """
        for subcommand in self.good_subcommands:
            cmd = 'simeon {sub} --help'.format(sub=subcommand)
            with self.subTest('Running {cmd}'.format(cmd=cmd)):
                self.proc = _run_command(cmd)
                self.assertEqual(self.proc.returncode, 0)
                self.tearDown()

    def test_bad_subcommands(self):
        """
        Test that mispelled subcommands should fail
        """
        for subcommand in self.bad_subcommands:
            cmd = 'simeon {sub} --help'.format(sub=subcommand)
            with self.subTest('Running {cmd}'.format(cmd=cmd)):
                self.proc = _run_command(cmd)
                self.assertNotEqual(self.proc.returncode, 0)
                self.tearDown()

    def test_global_options(self):
        """
        Test that the global options have not changed
        """
        for option in self.global_options:
            cmd = 'simeon {o} --help'.format(o=option)
            msg = 'Checking that global option {o} exists'.format(o=option)
            with self.subTest(msg):
                self.proc = _run_command(cmd)
                self.assertEqual(self.proc.returncode, 0)
                self.tearDown()


class SimeonGeoIPCLIIntegration(unittest.TestCase):
    """
    Test the simeon-geoip CLI tool
    """
    def setUp(self):
        self.this_dir = os.path.dirname(os.path.abspath(__file__))
        self.log_file = os.path.join(
            self.this_dir, 'fixtures', 'file.log'
        )
        self.config_file = os.path.join(
            self.this_dir, 'fixtures', 'config.ini'
        )
        self.un_file = os.path.join(
            self.this_dir, 'fixtures', 'un_data.csv'
        )
        self.output_file = os.path.join(
            self.this_dir, 'fixtures', 'geoip.json.gz'
        )
        self.global_options = [
            '--log-file {f}'.format(f=self.log_file),
            '--config-file {f}'.format(f=self.config_file),
            '--quiet',
        ]
        self.extract_good_options = [
            '--un-data {f}'.format(f=self.un_file),
            '--output {f}'.format(f=self.output_file),
            '--quiet', '--tracking-logs',
        ]
        self.merge_good_options = [
            '--geo-table dataset.table',
            '--project gcp-project-id',
            '--service-account-file saccount_file.json',
            '--column ip',
        ]
        self.proc = None

    def tearDown(self):
        if self.proc:
            streams = (self.proc.stdout, self.proc.stderr)
            for stream in streams:
                if stream:
                    stream.close()
        for file_ in (self.log_file, self.un_file, self.output_file):
            try:
                os.remove(file_)
            except OSError:
                continue

    def test_geoip_global_options(self):
        """
        Test that global options for simeon-geoip have not changed.
        """
        for option in self.global_options:
            cmd = 'simeon-geoip {o} --help'.format(o=option)
            msg = 'Checking that option {o} for simeon-geoip exists'.format(
                o=option
            )
            with self.subTest(msg):
                self.proc = _run_command(cmd)
                self.assertEqual(self.proc.returncode, 0)
                self.tearDown()

    def test_geoip_extract_options(self):
        """
        Test that options for simeon-geoip extract have not changed.
        """
        for option in self.global_options:
            cmd = 'simeon-geoip extract {o} --help'.format(o=option)
            msg = 'Checking that option {o} for extract exists'.format(
                o=option
            )
            with self.subTest(msg):
                self.proc = _run_command(cmd)
                self.assertEqual(self.proc.returncode, 0)
                self.tearDown()

    def test_geoip_merge_options(self):
        """
        Test that options for simeon-geoip merge have not changed.
        """
        for option in self.global_options:
            cmd = 'simeon-geoip merge {o} --help'.format(o=option)
            msg = 'Checking that option {o} for merge exists'.format(
                o=option
            )
            with self.subTest(msg):
                self.proc = _run_command(cmd)
                self.assertEqual(self.proc.returncode, 0)
                self.tearDown()


if __name__ == '__main__':
    unittest.main()



"""
Utility functions and classes associated with uploading data to GCP, so far.
"""
import json
import os
import re

from google.cloud import bigquery

from simeon.exceptions import (
    BigQueryNameException, MissingSchemaException
)


SEGMENTS = {
    'log': 'TRACKING-LOGS',
    'email': 'EMAIL',
    'sql': 'SQL',
    'rdx': 'RDX',
    'cold': 'COLD',
}
SCHEMA_DIR = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), 'schemas'
)
SCHEMAS = {}


def course_to_gcs_folder(course_id: str, file_type: str, bucket: str) -> str:
    """
    Use the given course ID to make a Google Cloud Storage path

    :type course_id: str
    :param course_id: edX course ID to format into a GCS path
    :type file_type: str
    :param file_type: One of sql, log, email, rdx
    :type bucket: str
    :param bucket: A GCS bucket name
    :rtype: str
    :return: A nicely formatted GCS path
    """
    segment = SEGMENTS.get(file_type)
    if not segment:
        raise ValueError(
            'file_type is not one of these: {ft}'.format(
                ft=', '.join(SEGMENTS)
            )
        )
    if not bucket.startswith('gs://'):
        bucket = 'gs://{b}'.format(b=bucket)
    dirname = course_id.replace('/', '__').replace('.', '_')
    return '{b}/{s}/{d}'.format(b=bucket, s=segment, d=dirname)


def local_to_gcs_path(fname: str, file_type: str, bucket: str) -> str:
    """
    Convert the local file name into a GCS path

    :type fname: str
    :param fname: A local file name
    :type file_type: str
    :param file_type: One of sql, log, email, rdx, cold
    :type bucket: str
    :param bucket: A GCS bucket name
    :rtype: str
    :return: A nicely formatted GCS path
    """
    segment = SEGMENTS.get(file_type)
    if not segment:
        msg = 'file_type is not one of these: {s}'
        raise ValueError(msg.format(s=', '.join(SEGMENTS)))
    if not bucket.startswith('gs://'):
        bucket = 'gs://{b}'.format(b=bucket)
    dname, bname = os.path.split(os.path.abspath(os.path.expanduser(fname)))
    if segment == 'COLD':
        gcs_file = bname
    else:
        gcs_file = '{d}/{f}'.format(
            d=os.path.basename(dname).replace('.', '_').replace('-', '_'),
            f=bname
        )
    return '{b}/{s}/{f}'.format(b=bucket, s=segment, f=gcs_file)


def course_to_bq_dataset(course_id: str, file_type: str, project: str) -> str:
    """
    Make a fully qualified BigQuery dataset name with the given info

    :type course_id: str
    :param course_id: edX course ID to format into a GCS path
    :type file_type: str
    :param file_type: One of sql, log, email, rdx
    :type project: str
    :param project: A GCP project ID
    :rtype: str
    :return: BigQuery dataset name with components separated by dots
    """
    if file_type not in SEGMENTS:
        msg = 'file_type is not one of these: {s}'
        raise ValueError(msg.format(s=', '.join(SEGMENTS)))
    if file_type in ('sql',):
        suffix = '_latest'
    elif file_type == 'log':
        suffix = '_logs'
    else:
        suffix = ''
    dataset = '{d}{s}'.format(
        d=course_id.replace('/', '__').replace('.', '_').replace('-', '_'),
        s=suffix
    )
    return '{p}.{d}'.format(p=project, d=dataset)


def local_to_bq_table(fname: str, file_type: str, project: str) -> str:
    """
    Use the given local file to make a fully qualified BigQuery table name

    :type fname: str
    :param fname: A local file name
    :type file_type: str
    :param file_type: One of sql, log, email, rdx
    :type project: str
    :param project: A GCP project ID
    :rtype: str
    :return: BigQuery dataset name with components separated by dots
    """
    if file_type not in SEGMENTS:
        raise ValueError(
            'file_type is not one of these: {s}'.format(
                s=', '.join(SEGMENTS)
            )
        )
    dname, bname = os.path.split(os.path.abspath(os.path.expanduser(fname)))
    if file_type in ('sql',):
        table = bname.split('.', 1)[0].replace('-', '_')
        suffix = '_latest'
    elif file_type in ('log',):
        suffix = '_logs'
        table = 'tracklog_' + ''.join(re.findall(r'\d+', bname))
    else:
        table = bname.split('.', 1)[0].replace('-', '_')
        suffix = ''
    dataset = os.path.basename(dname).replace('.', '_').replace('-', '_')
    if not table.replace('tracklog_', '') or not dataset:
        msg = (
            'A BigQuery table name for file type {t} '
            'could not be constructed with file name {f}. '
            'Please make sure that you are using the right file type.'
        )
        raise BigQueryNameException(msg.format(f=fname, t=file_type))
    return '{p}.{d}{s}.{t}'.format(
        d=dataset, s=suffix, p=project, t=table
    )


def dict_to_schema_field(schema_dict: dict):
    """
    Make a SchemaField from a schema directory

    :type schema_dict: dict
    :param schema_dict: One of the objects in the schema JSON file
    :rtype: bigquery.SchemaField
    :returns: A SchemaField matching the given dictionary's name, type, etc.
    """
    if schema_dict.get('field_type') != 'RECORD':
        return bigquery.SchemaField(**schema_dict)
    fields = []
    for subfield in schema_dict.get('fields', []):
        fields.append(dict_to_schema_field(subfield))
    schema_dict['fields'] = fields
    field = bigquery.SchemaField(**schema_dict)
    return field


def get_bq_schema(table: str, schema_dir: str=SCHEMA_DIR):
    """
    Given a bare table name (without leading project or dataset),
    make a list of bigquery.SchemaField objects to act as the table's schema.

    :type table: str
    :param table: A BigQuery (bare) table name
    :type schema_dir: str
    :param schema_dir: Directory where schema JSON file is looked up
    :rtype: Tuple[List[bigquery.SchemaField], str]
    :return: A 2-tuple with list of bigquery.SchemaField objects and
        a description text for the target table
    :raises: MissingSchemaException
    """
    bname = table.split('.')[-1]
    if all(k in bname for k in ('track', 'log')):
        bname = 'tracking_log'
    targets = ('{t}.json', 'schema_{t}.json')
    for target in targets:
        sfile = os.path.join(schema_dir, target.format(t=bname))
        if not os.path.exists(sfile):
            continue
        out = []
        with open(sfile) as fh:
            schema = json.load(fh)
            for field in schema.get(bname, []):
                out.append(dict_to_schema_field(field))
        if not out:
            msg = (
                'The schema file {f!r} does not contain an object '
                'with a name matching the given table {t}'
            )
            raise MissingSchemaException(msg.format(f=sfile, t=table))
        return out, schema.get('description')
    msg = 'No JSON schema file found for {t} in directory {d}'
    raise MissingSchemaException(msg.format(t=table, d=schema_dir))


def make_bq_load_config(
    table: str, schema_dir=SCHEMA_DIR, append: bool=False,
    create: bool=True, file_format: str='json', delim=',', max_bad_rows=0,
):
    """
    Make a bigquery.LoadJobConfig object and description of a table

    :type table: str
    :param table: Fully qualified table name
    :type schema_dir: str
    :param schema_dir: The directory where schema files live
    :type append: bool
    :param append: Whether to append the loaded to the table
    :type create: bool
    :param create: Whether to create the target table if it does not exist
    :type file_format: str
    :param file_format: One of sql, json, csv, txt
    :type delim: str
    :param delim: The delimiter of the file being loaded
    :rtype: Tuple[bigquery.LoadJobConfig, str]
    :return: A 2-tuple with a bigquery.LoadJobConfig object and
        a description text for the destination table
    """
    schema_dir = schema_dir or SCHEMA_DIR
    schema, desc = get_bq_schema(table, schema_dir=schema_dir)
    if 'json' in file_format.lower():
        format_ = bigquery.SourceFormat.NEWLINE_DELIMITED_JSON
        delim = None
        skips = None
    else:
        format_ = bigquery.SourceFormat.CSV
        skips = 1
    if create:
        create = bigquery.CreateDisposition.CREATE_IF_NEEDED
    else:
        create = bigquery.CreateDisposition.CREATE_NEVER
    if append:
        append = bigquery.WriteDisposition.WRITE_APPEND
    else:
        append = bigquery.WriteDisposition.WRITE_TRUNCATE
    if 'json' in file_format.lower():
        config = bigquery.LoadJobConfig(
            schema=schema, source_format=format_,
            create_disposition=create, write_disposition=append,
            max_bad_records=max_bad_rows, ignore_unknown_values=True,
            destination_table_description=desc,
        )
        return config, desc
    config = bigquery.LoadJobConfig(
        schema=schema, source_format=format_,
        create_disposition=create, write_disposition=append,
        field_delimiter=delim, skip_leading_rows=skips,
        max_bad_records=max_bad_rows, ignore_unknown_values=True,
        allow_quoted_newlines=True, destination_table_description=desc,
    )
    return config, desc


def make_bq_query_config(append: bool=False, plain=True, table=None):
    """
    Make a bigquery.QueryJobConfig object to tie to a query to be sent
    to BigQuery for secondary table generation

    :type append: bool
    :param append: Whether to append the loaded to the table
    :type plain: bool
    :param plain: Make an empty QueryJobConfig object
    :type table: Union[None, str]
    :param table: Fully qualified name of a destination table
    :rtype: bigquery.QueryJobConfig
    :returns: Make a bigquery.QueryJobConfig object
    """
    if plain:
        return bigquery.job.QueryJobConfig()
    if append:
        append = bigquery.WriteDisposition.WRITE_APPEND
    else:
        append = bigquery.WriteDisposition.WRITE_TRUNCATE
    config = bigquery.job.QueryJobConfig()
    if table is not None:
        config.destination = bigquery.Table.from_string(table)
    config.create_disposition = bigquery.CreateDisposition.CREATE_IF_NEEDED
    config.write_disposition = append
    config.allow_large_results = True
    config.use_legacy_sql = False
    return config


def sqlify_bq_field(field, named=True):
    """
    Convert a bigquery.SchemaField object into a DDL
    column definition.

    :type field: bigquery.SchemaField
    :param field: A SchemaField object to convert to a DDL statement
    :type named: bool
    :param named: Whether the returned statement start with the field name
    :rtype: str
    :returns: A SQL column's DDL statement
    """
    nullability = '' if field.is_nullable else 'NOT NULL'
    if 'INTEGER' in field.field_type:
        type_ = 'INT64'
    elif 'FLOAT' in field.field_type:
        type_ = 'FLOAT64'
    else:
        type_ = field.field_type
    if type_ != 'RECORD':
        if field.mode != 'REPEATED':
            return '{n} {t} {m} OPTIONS(description="""{d}""")'.format(
                n=field.name if named else '',
                t=type_,
                m=nullability,
                d=field.description or '',
            )
        else:
            return '{n} ARRAY<{t}> {m} OPTIONS(description="""{d}""")'.format(
                n=field.name if named else '',
                t=type_,
                m=nullability,
                d=field.description or '',
            )
    if type_ == 'RECORD' and field.mode != 'REPEATED':
        return '{n} STRUCT<{t}> {m} OPTIONS(description="""{d}"""'.format(
            t=',\n\t'.join(sqlify_bq_field(f) for f in field.fields),
            n=field.name if named else '',
            m=nullability,
            d=field.description or '',
        )
    field = bigquery.SchemaField(
        name=field.name, field_type=field.field_type,
        mode='NULLABLE' if field.is_nullable else 'REQUIRED',
        description=field.description, fields=field.fields,
    )
    return '{n} ARRAY<{t}> {m} OPTIONS(description="""{d}""")'.format(
        t=sqlify_bq_field(field, False).lstrip(' '),
        n=field.name,
        m=nullability,
        d=field.description or '',
    )



"""
Module of utilities to help with listing and downloading files from S3
"""
import json
import os
import re
import sys
import weakref
import zipfile
from datetime import datetime

import boto3 as boto

from simeon.exceptions import (
    AWSException, BlobDownloadError
)
from simeon.download.utilities import decrypt_files


BUCKETS = {
    'email': {
        'Bucket': '{bucket}',
        'Prefix': 'email-opt-in/email-opt-in-{org}-',
    },
    'sql': {
        'Bucket': '{bucket}',
        'Prefix': '{org}-',
    },
    'log': {
        'Bucket': '{bucket}',
        'Prefix': '{org}/{site}/events/{year}/{org}-{site}-events-',
    },
    'rdx': {
        'Bucket': '{bucket}',
        'Prefix': '{org}/rdx/{request}',
    },
}
BEGIN_DATE = '2012-09-01'
END_DATE = datetime.utcnow().strftime('%Y-%m-%d')
DATE_PATT = re.compile(r'\d{4}-\d{2}-\d{2}')
O2B_MAP = dict(
    key='name', last_modified='last_modified', size='size'
)


def make_s3_bucket(
    bucket, client_id=None, client_secret=None, session_token=None,
    profile_name=None,
):
    """
    Make a simple boto3 Bucket object pointing to S3
    """
    try:
        session = boto.Session(
            aws_access_key_id=client_id, aws_secret_access_key=client_secret,
            aws_session_token=session_token, profile_name=profile_name,
        )
        resource = session.resource('s3')
        return resource.Bucket(bucket)
    except Exception as excp:
        raise AWSException(excp)


def get_file_date(fname):
    """
    Get the date in the name of the S3 blob
    """
    fname = os.path.basename(fname)
    match = DATE_PATT.search(fname)
    if match:
        return match.group(0)
    return ''


class S3Blob():
    """
    A class to represent S3 blobs
    """
    def __init__(self, name, size, last_modified, bucket, local_name=None):
        """
        :type name: str
        :param name: Full path to the object without the bucket name
        :type size: int
        :param size: Size of the object in bytes
        :type last_modified: datetime
        :param last_modified: Last time the object was modified
        :type local_name: Union[None, str]
        :param local_name: Localized file name
        :type bucket: s3.Bucket
        :param bucket: The boto3.s3.Bucket object to tie to this blob
        """
        self.name = name
        self.size = size
        self.last_modified = last_modified
        self.bucket = weakref.proxy(bucket)
        self.bucket = bucket
        if not local_name:
            local_name = self._make_local(name)
        self.local_name = local_name

    @classmethod
    def from_prefix(cls, bucket, prefix):
        """
        Fetch a list of S3Blob objects from AWS whose names
        have the given prefix.

        :type bucket: s3.Bucket
        :param bucket: The boto3.s3.Bucket object to tie to this blob
        :type prefix: str
        :param prefix: A string with which to filter the list of objects
        :rtype: List[S3Blob]
        :return: A list of S3Blob objects
        :raises: AWSException
        """
        out = []
        maps = O2B_MAP
        try:
            matches = bucket.objects.filter(Prefix=prefix)
            for obj in matches:
                details = dict(
                    (v, getattr(obj, k, None)) for k, v in maps.items()
                )
                details['bucket'] = bucket
                out.append(cls(**details))
            return out
        except Exception as excp:
            raise AWSException('{e}'.format(e=excp)) from None

    @classmethod
    def from_info(cls, bucket, type_, date, org='mitx', site='edx'):
        """
        Make a list of blobs with the given parameters

        :type bucket: s3.Bucket
        :param bucket: The boto3.s3.Bucket object to tie to this blob
        :type type_: str
        :param type_: "sql" or "email" or "sql"
        :type date: Union[str, datetime]
        :param date: A datetime or str object for a threshold date
        :type org: str
        :param org: The org whose data will be fetched.
        :type site: str
        :param site: The site from which data were generated
        :rtype: List[S3Blob]
        :raises: AWSException
        """
        prefix = BUCKETS.get(type_, {}).get('Prefix')
        if not prefix:
            msg = (
                'The given file type, {t!r}, does not have any associated'
                ' AWS S3 information.'
            )
            raise AWSException(msg.format(t=type_)) from None
        if isinstance(date, datetime):
            date = date.strftime('%Y-%m-%d')
        year = date[:4]
        month = date[5:7]
        prefix = prefix.format(
            org=org, year=year, site=site, date=date, month=month
        )
        return cls.from_prefix(bucket, prefix)

    @staticmethod
    def _make_local(name):
        """
        Convert the given name into a local name for the file system
        """
        return os.path.join(*name.split('/'))

    def download_file(self, filename=None):
        """
        Download the S3Blob to the local file system
        and return the full path where the file is saved

        :type filename: Union[None, str]
        :param filename: Name of the output file
        :rtype: str
        :return: Returns the full path where the file is saved
        """
        if not filename:
            filename = self.local_name
        dirname, _ = os.path.split(filename)
        if dirname:
            os.makedirs(dirname, exist_ok=True)
        try:
            self.bucket.download_file(self.name, filename)
        except Exception as excp:
            msg = 'Failed to download blob {n}: {e}'
            raise BlobDownloadError(msg.format(n=self.name, e=excp))
        return filename

    def __repr__(self):
        return "Name: {n} - Size: {s} - Last Modified: {m}".format(
            n=self.name, s=self.size, m=self.last_modified
        )

    def to_json(self):
        """
        Jsonify the Blob
        """
        return json.dumps({
            'name': self.name, 'size': self.size,
            'last_modified': self.last_modified.strftime('%c %Z'),
        })



"""
Some utility functions for working with the downloaded data files
"""
import itertools as its
import gzip
import json
import math
import os
import re
import shlex
import subprocess as sb
import urllib.parse as urlparser
from datetime import datetime
from functools import lru_cache

from dateutil.parser import parse as parse_date

from simeon.exceptions import DecryptionError


COURSE_PATHS = [
    ('page',), ('event_type',), ('context', 'path'), ('name',)
]
MODULE_PATHS = [
    ('context', 'module', 'usage_key'), ('page',),
    ('context', 'path'), ('event', 'event_type'),
    ('name',), ('event', 'id'), ('event', 'problem_id'),
    ('event_type',)
]
SQL_FILE_EXTS = {
    '.failed': 3,
    '.gz': 3,
    '.json': 3,
    '.mongo': 1,
    '.sql': 3,
}


def _extract_values(record, paths):
    """
    Given a list of JSON paths (tuples), extract
    all the values associated with the given paths
    """
    for path in paths:
        subrec = record or {}
        start = path[:-1]
        end = path[-1]
        for k in start:
            subrec = record.get(k, {}) or {}
        if not isinstance(subrec, dict):
            continue
        yield subrec.get(end, '')


def _is_gpg_legacy():
    """
    Check if gpg is of a version below 2.1
    """
    p = sb.Popen(shlex.split('gpg --version'), stdout=sb.PIPE, stderr=sb.PIPE)
    if p.wait() != 0:
        raise DecryptionError(
            'The gpg command does not exist, '
            'or it is not properly configured.'
        )
    line = p.stdout.readline().decode('utf8', 'ignore').strip()
    version = line.split()[-1]
    return version < '2.1.0'


def decrypt_files(
    fnames, verbose=True, logger=None, timeout=None, keepfiles=False
):
    """
    Decrypt the given file with gpg.
    This assumes that the gpg command
    is available in the SHELL running this script.

    :type fnames: Union[str, List]
    :param fnames: A file name or a list of file names to decrypt
    :type verbose: bool
    :param verbose: Print the command to be run
    :type logger: logging.Logger
    :param logger: A logging.Logger object to print the command with
    :type timeout: Union[int, None]
    :param timeout: Number of seconds to wait for the decryption to finish
    :type keepfiles: bool
    :param keepfiles: Whether or not to keep the encrypted files after decryption
    :rtype: bool
    :return: Returns True if the decryption does not fail
    :raises: DecryptionError
    """
    if isinstance(fnames, str):
        fnames = [fnames]
    cmd = (
        'gpg {v}--status-fd 2 --batch --yes {p}'
        '--decrypt-files {f}'
    )
    verbosity = '--verbose ' if verbose else ''
    pinentry = '' if _is_gpg_legacy() else '--pinentry error '
    cmd = cmd.format(f=' '.join(fnames), v=verbosity, p=pinentry)
    if verbose and logger is not None:
        logger.info('{m}...'.format(m=cmd[:200]))
    proc =  sb.Popen(shlex.split(cmd), stdout=sb.PIPE, stderr=sb.PIPE)
    if proc.wait(timeout=timeout) != 0:
        errs = []
        for line in proc.stderr:
            errs.append(line.decode('utf8', 'ignore').strip())
        msg = 'Failed to decrypt file names {f} with return code {rc}:\n{e}'
        raise DecryptionError(msg.format(f=' '.join(
            fnames), e='\n'.join(errs), rc=proc.returncode
        ))
    if not keepfiles:
        for file_ in fnames:
            try:
                os.unlink(file_)
            except OSError as excp:
                if logger:
                    msg = 'Failed to delete encrypted file {f}: {e}'
                    logger.warning(msg.format(f=file_, e=excp))
    if verbose:
        msgs = []
        # If stderr is needed, then add it to the tuple below
        for stream in (proc.stdout,):
            for line in stream:
                line = line.decode('utf8', 'ignore').strip()
                if line:
                    msgs.append(line)
        if msgs and logger:
            logger.warning('\n'.join(msgs))
    return True


def get_file_date(fname):
    """
    Extract the date in a file name and parse it into a datetime object

    :type fname: str
    :param fname: Some file name
    :rtype: Union[None, datetime]
    :return: Returns a datetime object or None
    """
    fname = os.path.basename(fname)
    try:
        return parse_date(re.search(r'\d{4}-\d{2}-\d{2}', fname).group(0))
    except:
        return None


def make_file_handle(fname: str, mode: str='wt', is_gzip: bool=False):
    """
    Create a file handle pointing the given file name.
    If the directory of the file does not exist, create it.

    :type fname: str
    :param fname: A file name whose handle needs to be created.
    :type mode: str
    :param mode: "a[bt]?" for append or "w[bt]?" for write
    :type is_gzip: bool
    :param is_gzip: Whether or not to open it as a gzip file handle
    :rtype: Union[TextIOWrapper, BufferedReader]
    """
    fname = os.path.expanduser(fname)
    dirname, _ = os.path.split(fname)
    if dirname:
        os.makedirs(dirname, exist_ok=True)
    if is_gzip:
        return gzip.open(fname, mode)
    return open(fname, mode)


@lru_cache(maxsize=None)
def get_sql_course_id(course_str: str) -> str:
    """
    Given a course ID string from the SQL files,
    pluck out of the actual course ID and format it as follows:
    ORG/COURSE_NUMBER/TERM

    :type course_str: str
    :param course_str: The course ID string from edX
    :rtype: str
    :return: Actual course ID and format it properly
    """
    return course_str.split(':')[-1].replace('+', '/')


def format_sql_filename(fname: str):
    """
    Reformat the given edX SQL encrypted file name into a name indicative
    of where the file should end up after the SQL archive is unpacked.
    site/folder/filename.ext.gext
    """
    if fname.endswith('/'):
        return None, None
    file_ = fname.replace('prod-edge', 'edge').replace('ora/', '')
    if fname.endswith('.gpg'):
        file_, _ = os.path.splitext(file_)
    dirname, bname = os.path.split(file_)
    _, ext = os.path.splitext(bname)
    limit = SQL_FILE_EXTS.get(ext)
    if limit is None:
        raise ValueError(
            '{f} has an expected extension. Expected are {x}'.format(
                f=fname, x=', '.join(SQL_FILE_EXTS)
            )
        )
    components = bname.rsplit('-', limit)
    if '.mongo' in bname:
        cid, out = components
        site, out, ending = out.replace('.mongo', ''), 'forum.mongo.gpg', ''
    else:
        cid, out, site, ending = components
    out = '{o}-{e}.gpg'.format(o=out, e=ending) if ending else out
    if 'ora/' in fname:
        out = os.path.join('ora', out)
    return fname, os.path.join(site, dirname, cid, out)


def get_course_id(record: dict, paths=COURSE_PATHS) -> str:
    """
    Given a JSON record, try getting the course_id out of it.

    :type record: dict
    :param record: A deserialized JSON record
    :type paths: Iterable[Iterable[str]]
    :param paths: Paths to follow to find a matching course ID string
    :rtype: str
    :return: A valid edX course ID or an empty string
    """
    course_id = record.get('course_id')
    if not course_id:
        course_id = record.get('context', {}).get('course_id') or ''
    if not course_id:
        for course_id in _extract_values(record, paths):
            course_id = (course_id or '')
            if not (course_id.count('/') or course_id.count('+')):
                continue
            course_id = urlparser.urlparse(course_id).path
            if course_id:
                break
    course_id = (course_id or '').split('courses/')[-1]
    if course_id.count('i4x:') or course_id.count('data:image'):
        segments = '/'.join(course_id.split(':', 1)[0].split('+')[:3])
    else:
        segments = '/'.join(course_id.split(':', 1)[-1].split('+')[:3])
    return '/'.join([s for s in segments.rstrip('/').split('/') if s][:3])


def get_module_id(record: dict, paths=MODULE_PATHS):
    """
    Get the module ID of the given record

    :type record: dict
    :param record: A deserialized JSON record
    :type paths: Iterable[Iterable[str]]
    :param paths: Paths to follow to find a matching course ID string
    :rtype: str
    :return: A valid edX course ID or an empty string
    """
    values = _extract_values(record, paths)
    for value in values:
        if not value:
            continue
        if not any(k in value for k in '/-+@'):
            continue
        if value.startswith('i4x://'):
            value = value.lstrip('i4x://')
        if not value:
            continue
        block = urlparser.urlparse(value).path
        block = block.split('course-v1:')[-1]
        block = block.split('courses/')[-1]
        segments = block.split(':', 1)[-1].split('+')
        segments = '/'.join(map(lambda s: s.split('@')[-1], segments))
        return '/'.join([s for s in segments.split('/') if s][:5])
    return None


@lru_cache(maxsize=None)
def make_tracklog_path(course_id: str, datestr: str, is_gzip=True) -> str:
    """
    Make a local file path name with the given course ID and datetime object

    :type course_id: str
    :param course_id: Properly formatted edX course ID
    :type datestr: str
    :param datestr: %Y-%m-%d formatted date associated with the tracking log
    :type is_gzip: bool
    :param is_gzip: Whether or not we're making a GZIP file path
    :rtype: str
    :return: A local FS file path
    """
    ext = '.gz' if is_gzip else ''
    segments = course_id.strip().split('/')[:3]
    if len(segments) < 3:
        return os.path.join(
            'UNKNOWN', 'tracklog-{ds}.json{x}'.format(ds=datestr, x=ext)
        )
    return os.path.join(
        '__'.join(segments).replace('.', '_'),
        'tracklog-{ds}.json{x}'.format(ds=datestr, x=ext)
    )


@lru_cache(maxsize=None)
def parse_mongo_tstamp(timestamp: str):
    """
    Try converting a MongoDB timestamp into a stringified datetime

    :type timestamp: str
    :param timestamp: String representing a timestamp
    :rtype: str
    :return: A formatted datetime
    """
    if not timestamp:
        return ''
    try:
        timestamp = int(timestamp)
    except (TypeError, ValueError):
        try:
            return str(parse_date(timestamp))
        except Exception:
            try:
                return str(parse_date(timestamp[:16]))
            except Exception:
                msg = '{t} is not a valid timestamp'.format(t=timestamp)
                raise ValueError(msg) from None
    try:
        return str(datetime.fromtimestamp(timestamp/1000.0))
    except Exception:
        msg = '{t} is not a valid timestamp'.format(t=timestamp)
        raise ValueError(msg) from None


def check_for_funny_keys(record, name='toplevel'):
    """
    I am quite frankly not sure what Ike is trying to do here,
    but there should be a better way.
    For now, though, we'll just have to make do.

    :type record: dict
    :param record: Dictionary whose values are modified
    :type name: str
    :param name: Name of the level of the dict
    :rtype: None
    :return: Modifies the record in place
    """
    for key in list(record):
        val = record[key]
        if key.startswith('i4x-') or key.startswith('xblock.'):
            return True
        if key[0] in '0123456789':
            return True
        if '-' in key or '.' in key:
            newkey = key.replace('-', '_').replace('.', '__')
            record[newkey] = val
            record.pop(key)
            key = newkey
        if isinstance(val, dict):
            ret = check_for_funny_keys(val, name + '/' + key)
            if ret:
                record[key] = json.dumps(val)
    return False


def stringify_dict(record, *keys):
    """
    Given a dictionary and some keys, JSON stringify
    the values at those keys in place.

    :type record: dict
    :param record: Dictionary whose values are modified
    :type keys: Iterable[str]
    :param keys: multiple args
    :rtype: None
    :return: Modifies the dict in place
    """
    for key in keys:
        if isinstance(key, (list, tuple)) and len(key) > 1:
            key, subkey, *_ = key
            if key not in record or subkey not in record[key]:
                continue
            record[key][subkey] = json.dumps(record[key][subkey])
        else:
            if key not in record:
                continue
            record[key] = json.dumps(record[key])


def move_unknown_fields_to_agent(record, *keys):
    """
    Move the values associated with the given keys
    into record['agent']

    :type record: dict
    :param record: Dictionary whose values are modified
    :type keys: Iterable[str]
    :param keys: multiple args
    :rtype: None
    :return: Modifies the record in place
    """
    agent = {'oldagent': record.get('agent', '')}
    for key in keys:
        if '.' in key:
            prefix, subkey = key.split('.', 1)
            if prefix in record:
                subrecord = record[prefix]
                if subkey in subrecord:
                    agent[key] = subrecord[subkey]
                    subrecord.pop(subkey)
        else:
            if key in record:
                agent[key] = record[key]
                record.pop(key)
    record['agent'] = json.dumps(agent)


def is_float(val):
    """
    Check that the string can be coerced into a float.
    """
    try:
        float(val)
        return True
    except (TypeError, ValueError):
        return False


def move_field_to_mongoid(record: dict, path: list):
    """
    Move the values associated with the given path
    into record['mongoid']

    :type record: dict
    :param record: Dictionary whose values are modified
    :type keys: Iterable[str]
    :param keys: A list of keys to traverse and move
    :rtype: None
    :return: Modifies the record in place
    """
    mongoid = record.get('mongoid')
    if not isinstance(mongoid, dict):
        mongoid = {'old_mongoid': mongoid}
    key = path[0]
    if len(path) == 1:
        if key in record:
            val = record.pop(key)
            mongoid[key] = val
            return
        return
    if key not in mongoid:
        mongoid[key] = {}
    return move_field_to_mongoid(record, path[1:])


def drop_empties(record, *keys):
    """
    Recursive drop keys whose corresponding values are empty
    from the given record.

    :type record: dict
    :param record: Dictionary whose values are modified
    :type keys: Iterable[str]
    :param keys: multiple args
    :rtype: None
    :return: Modifies the record in place
    """
    if not keys:
        return
    key = keys[0]
    if isinstance(record, dict) and key in record:
        if len(keys) == 1:
            if record[key] == '':
                record.pop(key)
        else:
            drop_empties(record[key], *keys[1:])


def rephrase_record(record: dict):
    """
    Update the given record in place. The purpose of this function
    is to turn this record into something with the same schema as that of
    the target BigQuery table.

    :type record: dict
    :param record: A deserialized JSON record
    :rtype: None
    :return: Nothing, but updates the given record in place
    """
    record['course_id'] = get_course_id(record)
    record['module_id'] = get_module_id(record)
    if 'event' not in record:
        record['event'] = {}
    event = record.get('event')
    try:
        if not isinstance(event, dict):
            event = json.loads(event)
        event_js = True
    except:
        event_js = False
    record['event'] = event
    record['event_js'] = event_js
    event = record.get('event')
    if event is not None:
        record['event'] = json.dumps(event)
    event_type = record.get('event_type', '')
    known_types = set([
        'play_video', 'seq_goto', 'seq_next', 'seq_prev',
        'seek_video', 'load_video', 'save_problem_success',
        'save_problem_fail', 'reset_problem_success',
        'reset_problem_fail', 'show_answer',
        'edx.course.enrollment.activated',
        'edx.course.enrollment.deactivated',
        'edx.course.enrollment.mode_changed',
        'edx.course.enrollment.upgrade.succeeded', 'speed_change_video',
        'problem_check', 'problem_save', 'problem_reset'
    ])
    if isinstance(event, dict):
        outs = ('video_embedded', 'harvardx.button', 'harvardx.')
        out_conds = not any(k in event_type for k in outs)
        in_conds = 'problem_' in event_type or event_type in known_types
        if in_conds and out_conds:
            record['event_struct'] = event
        else:
            record['event_struct'] = {
                'GET': json.dumps(event.get('GET')),
                'POST': json.dumps(event.get('POST')),
                'query': event.get('query')
            }
    else:
        if 'event_struct' in record:
            record.pop('event_struct')
    if '_id' in record:
        record['mongoid'] = record['_id']['$oid']
        record.pop('_id')
    if isinstance(event, dict):
        if 'POST' in event:
            event['POST'] = json.dumps(event['POST'])
        if 'GET' in event:
            event['GET'] = json.dumps(event['GET'])
        if 'child-id' in event:
            event['child_id'] = event['child-id']
            event.pop('child-id')

    problem_events = set([
        'problem_check', 'problem_save', 'problem_reset'
    ])
    if event_type in problem_events and record['event_source'] == 'browser':
        if isinstance(event, (str, list, tuple)):
            event = {'data': json.dumps(event)}
    if isinstance(event, (str, list, tuple)):
        event = {'data': json.dumps(event)}
    to_stringify = (
        ('state', 'input_state'), ('state', 'correct_map'),
        ('state', 'student_answers'),
        'correct_map', 'answers', 'submission', 'old_state',
        'new_state', 'permutation', 'options_selected', 'corrections',
    )
    if event is not None:
        stringify_dict(event, *to_stringify)
    context = record.get('context', {})
    stringify_dict(context, 'course_user_tags')
    mobile_api_context_fields = [
        'application', 'client', 'received_at', 'component',
        'open_in_browser_url', 'module.usage_key',
        'module.original_usage_version', 'module.original_usage_key',
        'asides',
    ]
    move_unknown_fields_to_agent(context, *mobile_api_context_fields)
    mongo_paths = [
        ['referer'], ['accept_language'],
        ['event_struct', 'requested_skip_interval'],
        ['event_struct', 'submitted_answer'],
        ['event_struct', 'num_attempts'], ['event_struct', 'task_id'],
        ['event_struct', 'content'], ['nonInteraction'], ['label'],
        ['event_struct', 'widget_placement'], ['event_struct', 'tab_count'],
        ['event_struct', 'current_tab'], ['event_struct', 'target_tab'],
        ['event_struct', 'state', 'has_saved_answers'], ['context', 'label'],
        ['roles'], ['environment'], ['minion_id'],
        ['event_struct', 'duration'], ['event_struct', 'play_medium']
    ]
    for path in mongo_paths:
        move_field_to_mongoid(record, path)
    drop_empties(record, 'context', 'user_id')
    record.pop('event_js', '')
    if record.get('event_type') == 'speed_change_video':
        speed = record.get('event_struct', {}).get('new_speed')
        if is_float(speed):
            if math.isnan(float(speed)):
                record['event_struct'].pop('new_speed', '')
    check_for_funny_keys(record)



"""
Test the report package
"""
import glob
import json
import os
import unittest

import simeon.report.utilities as rutils


class TestReportUtilities(unittest.TestCase):
    """
    Test the utility and report generating functions
    in the simeon.report package
    """
    def setUp(self):
        self.blocks = [
            (
                "/courses/course-v1:MITx+0.502x+3T2019/"
                "xblock/block-v1:MITx+0.502x+3T2019+type@"
                "poll+block@d9c185016d32493e82239a541f7a274c/"
                "handler/student_voted"
            ),
            (
                "/courses/course-v1:MITx+0.502x+3T2019/xblock/"
                "block-v1:MITx+0.502x+3T2019+type@video+block@"
                "9636c575fdcf48e5b4dd54abbbd4ed28/handler_noauth/"
                "transcript/download"
            )
        ]
        self.courses = [
            'MITx/0.502x/3T2019', 'MITx/0.502x/3T2019',
        ]
        self.modules = [
            'MITx/0.502x/3T2019/poll/d9c185016d32493e82239a541f7a274c',
            'MITx/0.502x/3T2019/video/9636c575fdcf48e5b4dd54abbbd4ed28'
        ]
        self.fixtures_dir = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            'fixtures',
        )
        self.records = [
            {},
            {'user_id': 123, 'username': 'helloworld'},
            {'user_id': 321,},
        ]
        self.extra_records = [
            {
                'actor_name': 'Nova Jovic',
                'name': {
                    'first': 'Mella',
                    'last': 'Jameel'
                },
            },
            {
                'friends': ['Dior', 'Chanel'],
                'name': 'Cher',
            },
            {},
            {'user_id': 123, 'username': '123'},
        ]
        self.sfiles = [
            'schema_user_info_combo.json', 'schema_course_axis.json'
        ]
        self.schemas = []
        for f in self.sfiles:
            tbl = f.replace('schema_', '').replace('.json', '')
            f = os.path.join(rutils.SCHEMA_DIR, f)
            with open(f) as fh:
                self.schemas.append(json.load(fh).get(tbl))

    def tearDown(self):
        files = glob.glob(os.path.join(self.fixtures_dir, '*'))
        for file_ in files:
            try:
                os.remove(file_)
            except:
                continue
    
    def test_check_record_schema(self):
        """
        Test the check_record_schema function
        with records missing various fields
        """
        for record in self.records:
            msg = (
                'Test check_record_schema with {r} and coerce=True'
                ' and the schema files {f}'
            )
            with self.subTest(msg.format(r=record, f=', '.join(self.sfiles))):
                for schema in self.schemas:
                    copied = record.copy()
                    rutils.check_record_schema(
                        copied, schema, True, True
                    )
                    self.assertTrue(len(copied) > len(record))
            msg = (
                'Test check_record_schema with {r} and coerce=False'
                ' and the schema files for {f}'
            )
            with self.subTest(msg.format(r=record, f=', '.join(self.sfiles))):
                with self.assertRaises(rutils.SchemaMismatchException):
                    for schema in self.schemas:
                        updated = rutils.check_record_schema(
                            record.copy(), schema, False, True
                        )
    
    def test_drop_extra_keys(self):
        """
        Test the drop_extra_keys function with records
        that have extra fields.
        """
        for record in self.extra_records:
            msg = (
                'Test drop_extra_keys with {r} '
                ' and the schema files {f}'
            )
            with self.subTest(msg.format(r=record, f=', '.join(self.sfiles))):
                for schema in self.schemas:
                    copied = record.copy()
                    rutils.drop_extra_keys(copied, schema)
                    self.assertTrue(len(record) >= len(copied))
    
    def test_missing_extract_table_query(self):
        """
        Test the extract_table_query function for a table
        that does not exist.
        """
        with self.assertRaises(rutils.MissingQueryFileException):
            rutils.extract_table_query(
                'Idonotexistfornowbutmaybesomeday',
                rutils.QUERY_DIR,
            )
    
    def test_good_extract_table_query(self):
        """
        Test the extract_table_query function for person_course.
        """
        results = rutils.extract_table_query(
            'person_course',
            rutils.QUERY_DIR,
        )
        self.assertTrue(all(results))

    def test_make_sql_tables(self):
        """
        Test the make_sql_tables function with a directory with no files
        """
        with self.assertRaises(rutils.MissingFileException):
            rutils.make_sql_tables_seq(self.fixtures_dir, fail_fast=True)
    
    def test_make_course_axis(self):
        """
        Test the make_course_axis function inside simeon.report.utilities
        with an empty directory
        """
        with self.assertRaises(OSError):
            rutils.make_course_axis(self.fixtures_dir)

    def test_make_forum_table(self):
        """
        Test the make_forum_table function inside simeon.report.utilities
        with an empty directory
        """
        with self.assertRaises(OSError):
            rutils.make_forum_table(self.fixtures_dir)
    
    def test_make_grades_persistent(self):
        """
        Test the make_grades_persistent function inside simeon.report.utilities
        with an empty directory
        """
        with self.assertRaises(OSError):
            rutils.make_grades_persistent(self.fixtures_dir)

    def test_make_grading_policy(self):
        """
        Test the make_grading_policy function inside simeon.report.utilities
        with an empty directory
        """
        with self.assertRaises(OSError):
            rutils.make_grading_policy(self.fixtures_dir)

    def test_make_roles_table(self):
        """
        Test the make_roles_table function inside simeon.report.utilities
        with an empty directory
        """
        with self.assertRaises(OSError):
            rutils.make_roles_table(self.fixtures_dir)

    def test_make_student_module(self):
        """
        Test the make_student_module function inside simeon.report.utilities
        with an empty directory
        """
        with self.assertRaises(OSError):
            rutils.make_student_module(self.fixtures_dir)

    def test_make_user_info_combo(self):
        """
        Test the make_user_info_combo function inside simeon.report.utilities
        with an empty directory
        """
        with self.assertRaises(OSError):
            rutils.make_user_info_combo(self.fixtures_dir)

    def test_module_from_block(self):
        """
        Test the module_from_block function
        """
        for block, module in zip(self.blocks, self.modules):
            msg = 'Testing module_from_block with {b}'
            with self.subTest(msg.format(b=block)):
                self.assertEqual(
                    rutils.module_from_block(block), module
                )

    def test_course_from_block(self):
        """
        Test the course_from_block function
        """
        for block, course in zip(self.blocks, self.courses):
            msg = 'Testing course_from_block with {b}'
            with self.subTest(msg.format(b=block)):
                self.assertEqual(
                    rutils.course_from_block(block), course
                )
